"""
NRV-Cellular Level postprocessing.
.. warning::

    CL_postprocessing functions are deprecated. Use axon_results and associated methods instead.
"""

import faulthandler
from collections.abc import Iterable

import numpy as np
#from numba import jit
import matplotlib.pyplot as plt
from scipy import signal

from ..backend._file_handler import is_iterable, json_dump, json_load
from ..backend._log_interface import pass_info, rise_warning, rise_error
from ..backend._NRV_Class import load_any
from ..backend._file_handler import rmv_ext
from ..utils._units import MHz
from ..utils._misc import distance_point2line
from ..nmod._unmyelinated import unmyelinated
from ..nmod._myelinated import myelinated
from ..nmod.results._axons_results import axon_results


# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


##############################################
## HANDLE THE SIMULATION RESULT DICTIONNARY ##
##############################################
def remove_key(my_dict, key, verbose=False):
    """
    Remove an item from a dictionary, usefull before saving files, as some results maybe heavy and are potentially useless after some steps of postprocessing.

    Parameters
    ----------
    my_dict : dictionary
        dictionary where an item should be deleted
    key     : str
        name of the key to delete
    """
    
    rise_warning(
        "DeprecationWarning: ",
        "remove_key is obsolete use method from axon_result objects instead"
    )
    # if isinstance(key, Iterable):
    #    for k in key:
    #        del my_dict[k]
    # else:
    del my_dict[key]
    pass_info("removed the following key from results: ", key, verbose=verbose)


def remove_non_NoR_zones(my_dict, key):
    """
    Automatically remove values out of nodes of Ranvier for membrane voltage and associated quantities.
    This function is helpfull for large simulation before saving results

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be cleaned
    key     : str
        name of the key to clean
    """

    rise_warning(
            "DeprecationWarning: ",
            "remove_non_NoR_zones is obsolete use method from axon_result objects instead"
        )
    if "V_mem" in key:
        if my_dict["Axon_type"] == "Myelinated":
            new_entry = []
            for i in my_dict["Nodes_of_Ranvier_indexes"]:
                new_entry.append(my_dict[key][i, :])
            my_dict[key] = np.asarray(new_entry)
        else:
            rise_warning(
                "Warning, remove_non_NoR_zones only applicable to Myelinated axons"
            )
    else:
        rise_warning(
            "Warning, remove_non_NoR_zones only applicable to membrane voltage or current"
        )


def generate_axon_from_results(results):
    """
    generate a blank axon with the same dimensions as in the results

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    ax              : unmyelinated or myelinated
        blank axon with the same dimensions as in the results
    """

    # rise_warning(
    #     "DeprecationWarning: ",
    #     "generate_axon_from_results property is obsolete use method from axon_result objects instead"
    # )
    if not results["myelinated"]:
        ax = unmyelinated(**results)
    else:
        ax = myelinated(**results)
    return ax


############################
## AXON SIGNAL PROCESSING ##
############################
def filter_freq(my_dict, my_key, freq, Q=10):
    """
    Basic Filtering of quantities. This function design a notch filter (scipy IIR-notch).
    Adds an item to the specified dictionary, with the key termination '_filtered' concatenated to the original key.

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be filtered
    key     : str
        name of the key to filter
    freq    : float or array, list, np.array
        frequecy or list of frequencies to filter in kHz, as time is defined in ms in NRV2.
        If multiple frequencies, they are filtered sequencially, with as may filters as frequencies, in the specified order
    Q       : float
        quality factor of the filter, by default set to 10
    """

    rise_warning(
        "DeprecationWarning: ",
        "filter_freq is obsolete use method from axon_result objects instead"
    )
    if isinstance(freq, Iterable):
        f0 = np.asarray(freq)
    else:
        f0 = freq
    if my_dict["dt"] == 0:
        rise_warning(
            "Warning: filtering aborted, variable time step used for differential equation solving"
        )
        return False
    else:
        fs = 1 / my_dict["dt"]
        if isinstance(f0, Iterable):
            new_sig = np.zeros(my_dict[my_key].shape)
            for k in range(len(my_dict[my_key])):
                new_sig[k, :] = my_dict[my_key][k]
                for f in f0:
                    b_notch, a_notch = signal.iirnotch(f, Q, fs)
                    new_sig[k, :] = signal.lfilter(b_notch, a_notch, new_sig[k, :][k])
        else:
            ##  NOTCH at the stimulation frequency
            b_notch, a_notch = signal.iirnotch(f0, Q, fs)
            new_sig = np.zeros(my_dict[my_key].shape)
            for k in range(len(my_dict[my_key])):
                new_sig[k, :] = signal.lfilter(b_notch, a_notch, my_dict[my_key][k])
        my_dict[my_key + "_filtered"] = new_sig


def rasterize(
    my_dict, my_key, t_start=0, t_stop=0, t_min_spike=0.1, t_refractory=2, threshold=0
):
    """
    Rasterize a membrane potential (or filtered or any quantity processed from membrane voltage), with spike detection.
    This function adds 4 items to the dictionnary, with the key termination '_raster_position', '_raster_x_position', '_raster_time_index', '_raster_time' concatenated to the original key.
    These keys correspond to:
    _raster_position    : spike position as the indice of the original key
    _raster_x_position  : spike position as geometrical position in um
    _raster_time_index  : spike time as the indice of the original key
    _raster_time        : spike time as ms

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be rasterized
    key     : str
        name of the key to rasterize
    t_start         : float
        time at which the spike detection should start, in ms. By default 0
    t_stop          : float
        maximum time to apply spike detection, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    t_min_spike     : float
        minimum duration of a spike over its threshold, in ms. By default set to 0.1 ms
    t_refractory    : float
        refractory period for a spike, in ms. By default set to 2 ms.
    threshold       : float
        threshold for spike dection, in mV. If 0 is specified the threshold associated with the axon is automatically chosen. By default set to 0.
        Note that if a 0 value is wanted as threshold, a insignificat value (eg. 1e-12) should be specified.
    """

    rise_warning(
        "DeprecationWarning: ",
        "rasterize is obsolete use method from axon_result objects instead"
    )
    if t_stop == 0:
        t_stop = int(my_dict["t_sim"] / my_dict["dt"])
    else:
        t_stop = int(t_stop / my_dict["dt"])
    if threshold == 0:
        thr = my_dict["threshold"]
    else:
        thr = threshold
    ## selecting the list of position considering what has been recorded
    if my_dict["myelinated"] == True:
        if my_dict["rec"] == "all":
            list_to_parse = my_dict["node_index"]
            x = my_dict["x"]
        else:
            list_to_parse = np.arange(len(my_dict["x_rec"]))  # my_dict[my_key]
            x = my_dict["x_rec"]
    else:
        list_to_parse = np.arange(len(my_dict["x_rec"]))  # my_dict[my_key]
        x = my_dict["x_rec"]
    # spike detection
    (
        my_dict[my_key + "_raster_position"],
        my_dict[my_key + "_raster_x_position"],
        my_dict[my_key + "_raster_time_index"],
        my_dict[my_key + "_raster_time"],
    ) = AP_detection(
        my_dict[my_key],
        my_dict["t"],
        x,
        list_to_parse,
        thr,
        my_dict["dt"],
        t_start,
        t_stop,
        t_refractory,
        t_min_spike,
    )


#@jit(nopython=True, fastmath=True)
def AP_detection(
    Voltage, t, x, list_to_parse, thr, dt, t_start, t_stop, t_refractory, t_min_spike
):
    """
    Internal use only, spike detection just in time compiled to speed up the process
    """

    rise_warning(
        "DeprecationWarning: ",
        "AP_detection is obsolete use method from axon_result objects instead"
    )
    raster_position = []
    raster_x_position = []
    raster_time_index = []
    raster_time = []
    # parsing to find spikes
    for i in list_to_parse:
        t_last_spike = t_start - t_refractory
        for j in range(int(t_start * (1 / dt)), t_stop):
            if (
                Voltage[i][j] <= thr
                and Voltage[i][j + 1] >= thr
                and Voltage[i][min((j + int(t_min_spike * (1 / dt))), t_stop)] >= thr
                and (j * dt - t_last_spike) > t_refractory
            ):  # 1st line: threshold crossing, 2nd: minimum time above threshold,3rd: refractory period
                # there was a spike, get time and position
                raster_position.append(i)
                raster_x_position.append(x[i])
                raster_time_index.append(j)
                raster_time.append(t[j])
                # memorize the time in ms, to evaluate refractory period
                t_last_spike = j * dt
    # return results
    return (
        np.asarray(raster_position),
        np.asarray(raster_x_position),
        np.asarray(raster_time_index),
        np.asarray(raster_time),
    )



def speed(my_dict, position_key=None, t_start=0, t_stop=0, x_start=0, x_stop=0):
    """
    Compute the velocity of a spike from rasterized data in a dictionary. The speed can be either positive or negative depending on the propagation direction.

    Parameters
    ----------
    my_dict     : dictionary
        dictionary where the quantity should be rasterized
    key         : str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start     : float
        time at which the spikes are processed, in ms. By default 0
    t_stop      : float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    x_start     : float
        minimum position for spike processing, in um. By default set to 0.
    x_stop      : float
        maximum position for spike processing, in um. If 0 is specified, spikes are processed on the full axon length . By default set to 0.

    Returns
    -------
    speed   : float
        velocity.

    Note
    ----
    the velocity is computed with first and last occurance of a spike, be careful specifying the computation window if multiple spikes. This function will not see velocity variation.
    """

    rise_warning(
        "DeprecationWarning: ",
        "speed is obsolete use method from axon_result objects instead"
    )
    # define max timing if not already defined
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    if t_start == 0:
        if "intra_stim_starts" in my_dict and my_dict["intra_stim_starts"] != []:
            t_start = my_dict["intra_stim_starts"][0]
    if x_start == 0:
        x_stop = my_dict["L"]
    elif x_stop == 0:
        x_start = my_dict["L"]
    # find the best raster plot
    if position_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    else:
        good_key_prefix = position_key
    # get data only in time windows
    sup_time_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_time_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
    sup_position_indexes = np.where(
        my_dict[good_key_prefix + "_x_position"][good_indexes_time] >= x_start
    )
    inf_position_indexes = np.where(
        my_dict[good_key_prefix + "_x_position"][good_indexes_time] <= x_stop
    )
    good_indexes_position = np.intersect1d(sup_position_indexes, inf_position_indexes)
    good_indexes = np.intersect1d(good_indexes_position, good_indexes_time)
    good_spike_times = my_dict[good_key_prefix + "_time"][good_indexes]
    good_spike_positions = my_dict[good_key_prefix + "_x_position"][good_indexes]
    max_time = np.argmax(good_spike_times)
    min_time = np.argmin(good_spike_times)
    speed = (
        (good_spike_positions[max_time] - good_spike_positions[min_time])
        * 10**-3
        / (good_spike_times[max_time] - good_spike_times[min_time])
    )
    return speed


def block(my_dict, position_key=None, t_start=0, t_stop=0):
    """
    check if an axon is blocked or not. The simulation has to include the test spike. This function will look for the test spike initiation and check the propagation

    Parameters
    ----------
    my_dict 	: dictionary
        dictionary where the quantity should be rasterized
    key 		: str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start		: float
        time at which the test spikes can occur, in ms. By default 0
    t_stop		: float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.

    Returns
    -------
    flag 	: bool or None
        True if the axon is blocked, False if not blocked and None if the test spike does not cross the stimulation near point in the simulation (no possibility to check for the axon state)
    """

    rise_warning(
        "DeprecationWarning: ",
        "block is obsolete use method from axon_result objects instead"
    )
    position_max = 0
    blocked_spike_positionlist = []
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    if t_start == 0:
        if "intra_stim_starts" in my_dict and my_dict["intra_stim_starts"] != []:
            t_start = my_dict["intra_stim_starts"][0]

    if position_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    sup_time_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_time_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
    good_spike_times = my_dict[good_key_prefix + "_time"][good_indexes_time]
    blocked_spike_positionlist = my_dict[good_key_prefix + "_x_position"][
        good_indexes_time
    ]
    if len(blocked_spike_positionlist) == 0:
        return None
    if "intra_stim_positions" in my_dict:
        if my_dict["intra_stim_positions"] < my_dict["extracellular_electrode_x"]:
            position_max = max_spike_position(
                blocked_spike_positionlist, position_max, spike_begin="down"
            )

            if blocked_spike_positionlist[position_max] < 9.0 / 10 * my_dict["L"]:
                return True
            else:
                for i in range(position_max - 1):
                    if (
                        blocked_spike_positionlist[i + 1]
                        - blocked_spike_positionlist[i]
                        > my_dict["L"] / 5
                    ):
                        return True
                else:
                    return False
        else:
            position_max = max_spike_position(
                blocked_spike_positionlist, position_max, spike_begin="up"
            )
            if min(blocked_spike_positionlist) > 1.0 / 10 * my_dict["L"]:
                return True
            else:
                for i in range(position_max - 1):
                    if (
                        blocked_spike_positionlist[i]
                        - blocked_spike_positionlist[i + 1]
                        > my_dict["L"] / 5
                    ):
                        return True
                else:
                    return False
    else:
        pass_info("intra_stim_positions is not in dictionnary")


def max_spike_position(blocked_spike_positionlist, position_max, spike_begin="down"):
    rise_warning(
        "DeprecationWarning: ",
        "max_spike_position is obsolete use method from axon_result objects instead"
    )
    if spike_begin == "down":
        while blocked_spike_positionlist[
            position_max + 1
        ] >= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max
    else:
        while blocked_spike_positionlist[
            position_max + 1
        ] <= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max


#@jit(nopython=True, fastmath=True)
def count_spike(onset_position):
    """
    spike counting, just in time compiled. For internal use only.
    """

    rise_warning(
        "DeprecationWarning: ",
        "count_spike is obsolete use method from axon_result objects instead"
    )
    if len(onset_position) == 0:
        spike_number = 0
        return 0
    else:
        spike_number = 1
        for i in range(len(onset_position) - 1):
            if onset_position[i] == min(onset_position):
                if onset_position[i] == onset_position[i + 1]:
                    spike_number = spike_number + 1
    return spike_number


def check_test_AP(results_sim):
    """
    Check if a test AP is correctely triggered during an axon simulation and if so return the\
    trigger time

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    test_AP     : float or None
        time in ms of the first test AP during the simulation. None if no test AP found
    """

    rise_warning(
        "DeprecationWarning: ",
        "check_test_AP is obsolete use method from axon_result objects instead"
    )
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)
    if "intra_stim_starts" not in results_sim:
        return None
    else:
        mask = False
        test_AP = results_sim["intra_stim_starts"]
        if len(test_AP):
            if is_iterable(test_AP):
                test_AP = test_AP[0]
            i_first_pos = np.where(results_sim["V_mem_raster_x_position"] == 0)
            for i in i_first_pos[0]:
                if (
                    results_sim["V_mem_raster_time"][i] >= test_AP - 0.01
                    and results_sim["V_mem_raster_time"][i] <= test_AP + 0.7
                ):
                    mask = True
            if not mask:
                test_AP = None
        else:
            test_AP = None
        return test_AP


def detect_start_extrastim(results_sim, threshold=None):
    """
    Returns the starting time of extracellular stimulation from axon simulation results

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method
    threshold       : float
        Current threshold (uA) to consider the stimulation started, if None take the time of second point of the\
        stimulation, by default None

    Returns
    -------
    t_start     : list or float or None
        list of stimulation starting time in ms, float if only one stimulation and None if no stimulation
    """

    rise_warning(
        "DeprecationWarning: ",
        "detect_start_extrastim is obsolete use method from axon_result objects instead"
    )
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    t_start = []
    if "extracellular_stimuli" in results_sim:
        for i in range(len(results_sim["extracellular_stimuli"])):
            s = results_sim["extracellular_stimuli"][i]
            t = results_sim["extracellular_stimuli_t"][i]
            if threshold is None:
                t_start += [t[1]]
            else:
                for i in reversed(range(len(s))):
                    if abs(s[i]) > threshold:
                        t_start += [t[i]]
    if len(t_start) == 0:
        t_start = None
    elif len(t_start) == 1:
        t_start = t_start[0]
    return t_start


def extra_stim_properties(results_sim):
    """
    Return elect caracteristics (blocked, Onset response, ...)

    Parameters
    ----------
    results_sim       : dict or str
        simulation results dictionary or path and name of the file containing it
    Returns
    -------
    electrode       : dict
        dictonry containing the position (x, y, z), the stimulation start (ms) and maximum value (uA)
        for each electrodes
    """

    rise_warning(
        "DeprecationWarning: ",
        "extra_stim_properties is obsolete use method from axon_result objects instead"
    )
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    electrode = {}

    if "extracellular_electrode_y" in results_sim:
        electrode["x"] = results_sim["extracellular_electrode_x"]
        electrode["y"] = results_sim["extracellular_electrode_y"]
        electrode["z"] = results_sim["extracellular_electrode_z"]
        electrode["start"] = detect_start_extrastim(results_sim)
        electrode["max amplitude"] = [
            max(s) for s in results_sim["extracellular_stimuli"]
        ]

    return electrode


def axon_state(results_sim, save=False, saving_file="axon_state.json"):
    """
    Return axon caracteristics (blocked, Onset response, ...) from axon simulation results

    Parameters
    ----------
    results_sim       : dict or str
        simulation results dictionary or path and name of the saving file
    save        : bool
        if True save result in json file
    saving_file : str
        if save is True path and name of the saving file

    Returns
    -------
    axon_state       : dict
        dictionary containing axon caracteristics
    """

    rise_warning(
        "DeprecationWarning: ",
        "axon_state is obsolete use method from axon_result objects instead"
    )
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    ID = results_sim["ID"]

    # Axon parameter
    parameters = {}

    if "diameter" in results_sim:
        parameters["diameter"] = results_sim["diameter"]

    if "myelinated" in results_sim and results_sim["myelinated"]:
        parameters["node"] = len(results_sim["x_nodes"])

    if (
        "extracellular_electrode_y" in results_sim
        and len(results_sim["extracellular_electrode_y"]) == 1
    ):
        parameters["distance electrod"] = distance_point2line(
            results_sim["y"],
            results_sim["z"],
            results_sim["extracellular_electrode_y"][0],
            results_sim["extracellular_electrode_z"][0],
        )
        if results_sim["myelinated"]:
            x_elec = results_sim["extracellular_electrode_x"][0]
            elec_node = np.argmin(abs(results_sim["x_nodes"] - x_elec))
            elec_ali = (results_sim["x_nodes"][elec_node] - x_elec) / (
                results_sim["x_nodes"][1] - results_sim["x_nodes"][0]
            )
            parameters["electrod node"] = int(elec_node)
            parameters["electrod alignment"] = float(elec_ali)

    # Check Block
    test_AP = check_test_AP(results_sim)
    if test_AP is None:
        block_state = None
    else:
        if "extracellular_electrode_x" not in results_sim:
            results_sim["extracellular_electrode_x"] = 0
        block_state = block(
            results_sim, t_start=test_AP - 0.001
        )  # , t_stop=test_AP+1) # Gerer le delay
    # Check Onset Response

    onset_state = False
    t_start_stim = detect_start_extrastim(results_sim)

    pos = results_sim["V_mem_raster_position"]
    if results_sim["myelinated"]:
        M = len(["x_nodes"])
    else:
        M = len(["x_rec"])
    i_first_pos = np.where(pos == 0)
    i_last_pos = np.where(pos == M)

    # Count Onset response
    N_AP = (len(i_first_pos[0]) + len(i_last_pos[0])) / 2
    if test_AP is not None:
        if block_state:
            N_AP -= 0.5
        else:
            N_AP -= 1

    if N_AP > 0:
        onset_state = True

    axon_state = {
        "ID": ID,
        "parameters": parameters,
        "block_state": block_state,
        "onset_state": onset_state,
        "onset number": N_AP,
    }
    if save:
        save_axon_results_as_json(axon_state, saving_file)

    return axon_state


##############################
## Axon properties function ##
##############################
def get_index_myelinated_sequence(results, n):
    """
    Returns the sequence

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    n_center              : int
        value of the cutoff frequency of the axon's membrane
    """

    rise_warning(
        "DeprecationWarning: ",
        "get_index_myelinated_sequence is obsolete use method from axon_result objects instead"
    )
    if not results["myelinated"] or results["rec"] == "node":
        return "node"
    else:
        if n > len(results["x_rec"]):
            rise_warning("index not in axon")
        # +1 required because nbr of computation point = nbr seg/sec + 1
        # see if it's a bug
        Nseg_per_sec = results["Nseg_per_sec"] + 1
        N_sec_type = 11
        seq_types = results["axon_path_type"]
        if n == 0:
            return seq_types[0]
        else:
            return seq_types[((n - 1) // Nseg_per_sec) % N_sec_type]


def find_central_node_index(results:axon_results)->int:
    """
    Returns the index of the closer node from the center from a dictionnary results

    Parameters
    ----------
    axon_results     : dict
        results of axon.simulate method

    Returns
    -------
    n_center              : int
        value of the cutoff frequency of the axon's membrane
    """

    rise_warning(
        "DeprecationWarning: ",
        "find_central_node_index is obsolete use method from axon_result objects instead"
    )
    return results.find_central_index()


def compute_f_mem(results):
    """
    compute the cutoff frequency of the axon's membrane and add it to the simulation results dictionnary
    NB: The frequency is computed in [kHz]

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    f_mem              : np.ndarray
        value of the cutoff frequency of the axon's membrane
    """

    rise_warning(
        "DeprecationWarning: ",
        "compute_f_mem is obsolete use method from axon_result objects instead"
    )
    if "g_mem" not in results:
        rise_warning("f_mem cannot be computed computed without membrane conductivity")
        return None
    if "f_mem" not in results:
        ax = generate_axon_from_results(results)
        results["c_mem"] = ax.get_membrane_capacitance()
        del ax
    N_x, N_t = np.shape(results["g_mem"])
    f_mem = np.zeros((N_x, N_t))
    for i_t in range(N_t):
        f_mem[:, i_t] = results["g_mem"][:, i_t] / (2 * np.pi * results["c_mem"])

    # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
    # * [MHz] to convert to [kHz]
    results["f_mem"] = f_mem * MHz
    return results["f_mem"]


def get_myelin_properties(results):
    """
    compute the cutoff frequency of the axon's membrane and add it to the simulation results dictionnary
    NB: The frequency is computed in [kHz]

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    f_mem              : np.ndarray
        value of the cutoff frequency of the axon's membrane
    """

    rise_warning(
        "DeprecationWarning: ",
        "get_myelin_properties is obsolete use method from axon_result objects instead"
    )
    if not results["myelinated"] or results["rec"] == "node":
        rise_warning("No myelin in the axon simulated, None returned")
        return None

    ax = generate_axon_from_results(results)
    results["g_mye"] = ax.get_myelin_conductance()
    results["c_mye"] = ax.get_myelin_capacitance()
    results["f_mye"] = results["g_mye"] / (2 * np.pi * results["c_mye"])

    # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
    # * [MHz] to convert to [kHz]
    results["f_mem"] *= MHz
    return results["f_mem"]


#############################
## VISUALIZATION FUNCTIONS ##
#############################
def plot_Nav_states(ax, values, title=""):
    """
    Plot the state machine for kinetic (Markov) Nav 1.1 to 1.9 values

    Parameters
    ----------
    ax : matplotlib axis object
        axes of the figure to work on
    values : list, array, numpy array
        
    title : str, optional
        Title of the plot, by default ""
    """


    rise_warning(
        "DeprecationWarning: ",
        "plot_Nav_states is obsolete use method from axon_result objects instead"
    )
    states = [r"$I_1$", r"$I_2$", r"$C_1$", r"$C_2$", r"$O_1$", r"$O_2$"]

    X = [-1, -3, 0, 1, 0, 3]
    Y = [0, 0, 1, 0, -1, 0]
    c = ["r", "r", "b", "b", "g", "g"]

    ax.set_xlim(-3.4, 3.4)
    ax.set_ylim(-1.5, 1.5)
    for i in range(len(states)):
        ax.scatter(X[i], Y[i], s=300 + values[i] * 1450, c=c[i], alpha=0.4)
        ax.text(X[i], Y[i], states[i], ha="center", va="center")
    # paths
    ax.arrow(
        -2.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-2, 0.2, "$I_2I_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        -1.5, -0.03, -1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-2, -0.2, "$I_1I_2$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        -0.83, 0.25, 0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-0.9, 0.6, "$I_1C_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        -0.22,
        0.75,
        -0.5,
        -0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(-0.45, 0.25, "$C_1I_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        0.72, 0.25, -0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.9, 0.6, "$C_1C_2$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        0.33, 0.75, 0.5, -0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.45, 0.25, "$C_2C_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        0.83,
        -0.25,
        -0.5,
        -0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(0.9, -0.6, "$C_2O_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        0.22, -0.75, 0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.45, -0.25, "$O_1C_2$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        -0.33,
        -0.75,
        -0.5,
        0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(-0.9, -0.6, "$O_1I_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(1.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02)
    ax.text(2, 0.2, "$C_20_2$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        2.5, -0.03, -1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(2, -0.2, "$O_2C_2$", ha="center", va="center", alpha=0.4)
    # make axes to disappear
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(title)
    ax.axis("off")


################################
## usefull methods on results ##
################################

def default_PP (results:axon_results)->axon_results:
    """
    Default postprocessing function. Rasterize data and remove v_mem to alliviate RAM usage

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.

    Returns
    -------
    axon_results
        updated results of the axon simulation.
    """
    results.rasterize()
    results.remove_key("V_mem")
    return(results)

def rmv_keys(results:axon_results, keys_to_remove:str|set[str]={}, keys_to_keep:set[str]={})->axon_results:
    """
    remove most of the results key to save computing memory.

    Note
    ----
    Only the following keys kept by the fuction
     - `ID`
     - `L`
     - `V_mem_raster_position`
     - `V_mem_raster_x_position`
     - `V_mem_raster_time_index`
     - `V_mem_raster_time`
     - `myelinated`
     - `intra_stim_starts`
     - `intra_stim_positions`


    Warning
    -------
    This function removes a large amount of data from the `results` object. Some methods included in the results may not be used afterwards

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    keys_to_remove : str | list[str], optional
        key or set of key that should be removed, by default []
    keys_to_keep : str | list[str], optional
        If None only keys_to_remove are removed. Otherwise, all key exept those in this list are deleted, by default None.

    Returns
    -------
    axon_results
        updated results.
    """

    results.rasterize("V_mem")
    default_list_keys = {
    "ID",
    "L",
    "V_mem_raster_position",
    "V_mem_raster_x_position",
    "V_mem_raster_time_index",
    "V_mem_raster_time",
    "myelinated",
    "intra_stim_starts",
    "intra_stim_positions",
    "recorder"
    }

    default_list_keys.update(set(keys_to_keep))
    default_list_keys -= set(keys_to_remove)
    results.remove_key(keys_to_keep=default_list_keys)
    return results


def is_recruited(results:axon_results)->axon_results:
    """
    Evaluate if each fibre is recruited by a stimulation (see `axon_results.is_recruited`) and remove most of the `axon_results` keys to alliviate RAM usage.

    Note
    ----
    Only the following keys kept by the fuction
     - `ID`
     - `L`
     - `V_mem_raster_position`
     - `V_mem_raster_x_position`
     - `V_mem_raster_time_index`
     - `V_mem_raster_time`
     - `myelinated`
     - `y`
     - `z`
     - `diameter`
     - `intra_stim_starts`
     - `tstop`
     - `intra_stim_positions`
     - `extracellular_electrode_x`
     - `recruited`

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.

    Returns
    -------
    axon_results
        updated results.
    """
    results.is_recruited(vm_key="V_mem")

    # remove non nevessary data
    list_keys = {
    "ID",
    "L",
    "V_mem_raster_position",
    "V_mem_raster_x_position",
    "V_mem_raster_time_index",
    "V_mem_raster_time",
    "myelinated",
    "y",
    "z",
    "diameter",
    "intra_stim_starts",
    "tstop",
    "intra_stim_positions",
    "extracellular_electrode_x",
    "recruited",
    }
    results.remove_key(keys_to_keep=list_keys)
    return results

def is_blocked(results:axon_results, AP_start:float|None=None, freq:float|None=None, t_refractory:float=1)->axon_results:
    """
    Evaluate the impact od a blocking stimulation on axon (see `axon_results.block_summary`) and remove most of the `axon_results` keys to alliviate RAM usage.

    Note
    ----
    Only the following keys kept by the fuction
     - `ID`
     - `L`
     - `V_mem_raster_position`
     - `V_mem_raster_x_position`
     - `V_mem_raster_time_index`
     - `V_mem_raster_time`
     - `myelinated`
     - `y`
     - `z`
     - `diameter`
     - `intra_stim_starts`
     - `tstop`
     - `intra_stim_positions`
     - `extracellular_electrode_x`
     - `blocked`
     - `has_onset`
     - `n_onset`

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    save : bool, optional
        if true, the block status is saved as a line in a `.csv` file, by default False
    fdir : str, optional
        Path where the conductivity should be saved, the saving file will be `fdir+block_summary.csv`, by default ""

    Note
    ----
    If `save` is true, the data arranged as column of a same line in the following order:
     - `ID`
     - `y`
     - `z`
     - `diameter`
     - `myelinated`
     - `is_blocked`
     - `has_onset`
     - `n_onset`

    Returns
    -------
    axon_results
        updated results.
    """
    ## TO CHANGE WHEN is block is developped
    #results.axon_state(save=False)

    if AP_start is None :
        if "intra_stim_starts" in results and results["intra_stim_starts"] != []:
            AP_start = results["intra_stim_starts"][0]

    vm_key = "V_mem"
    if freq is not None:
        vm_key += "_filtered"

    results.block_summary(AP_start=AP_start, freq=freq, t_refractory=t_refractory)

    # remove non nevessary data
    list_keys = {
    "ID",
    "L",
    f"{vm_key}_raster_position",
    f"{vm_key}_raster_x_position",
    f"{vm_key}_raster_time_index",
    f"{vm_key}_raster_time",
    "myelinated",
    "y",
    "z",
    "diameter",
    "intra_stim_starts",
    "tstop",
    "intra_stim_positions",
    "extracellular_electrode_x",
    "is_blocked",
    "has_onset",
    "n_onset",
    }
    results.remove_key(keys_to_keep=list_keys)
    return results



def sample_keys(results:axon_results, keys_to_sample:str|set[str]={}, t_start_rec:float=0, t_stop_rec:float=-1, sample_dt:None|float=None, i_sampled_t:None|np.ndarray=None, x_bounds:None|float|tuple[float]=None, keys_to_remove:str|set[str]={}, keys_to_keep:set[str]={})->axon_results:
    """
    Undersample the membrane coductivity (``results["g_mem"]``) key and remove most of the `axon_results` keys to alliviate RAM usage.

    Note
    ----
    Only the following keys kept by the fuction
     - `g_mem`
     - `x_rec`
     - `rec`
     - `Nseg_per_sec`
     - `axon_path_type`
     - `t_sim`

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    t_start_rec : float, optional
        Lower time at whitch `g_mem` should be stored, by default 0
    t_stop_rec : float, optional
        Upper time at whitch `g_mem` should be stored, by default -1
    sample_dt : None | float, optional
        Time sample rate at which `g_mem` should be stored if None simulation dt is kept, by default None
    x_bounds : None | tuple[float], optional
        x-positions where to store `g_mem`, possible option:
         - float: The values of `g_mem` are only stored at the nearest position in `x_rec`.
         - tupple: `g_mem` values are stored for all positions included between the two boundaries.
         - None (default): `g_mem` values are stored for all positions.

    Warning
    -------
    ``sample_dt`` should be at multiple of the simulation ``dt`` to allow a correct undersampling. 
    If the not ``sample_dt`` choosen will be the closer multiple of ``dt``.

    Returns
    -------
    axon_results
        updated results.
    """
    if isinstance(keys_to_sample, str):
        keys_to_sample = {keys_to_sample}
    if len(set(keys_to_sample) - set(results.keys())):
        rise_error(set(keys_to_sample) - set(results.keys()), "keys are missing to apply postprocessing. Please check simulation parameters")
    else:
        # x - sampling array
        if x_bounds is None:
            I_x = np.arange(len(results["x_rec"]))
            x_bounds = (0, results["x_rec"][-1])
        elif np.iterable(x_bounds):
            I_x = np.argwhere((results["x_rec"]>x_bounds[0])&(results["x_rec"]<x_bounds[1]))[:,0]
        else:
            x_bounds = [x_bounds]
            I_x = np.array([np.argmin(abs(results["x_rec"]-x_bounds[0]))])
        N_x = len(I_x)

        # t - sampling array
        if i_sampled_t is not None:
            t_APs = i_sampled_t
        else:
            if t_stop_rec < 0:
                i_t_max = len(results["t"])
            else:
                i_t_max = np.argwhere(results["t"]<=t_stop_rec)[-1][0]
            if sample_dt is None:
                sample_dt=results.dt
            i_t_min = np.argwhere(results["t"]>=t_start_rec)[0][0]
             
            t_APs = [k for k in range(i_t_min,i_t_max)]
            t_APs = t_APs[::int(sample_dt/results.dt)]

        # Under sampling to reduce memory consumption
        results["x_rec"] = results["x_rec"][I_x] - x_bounds[0]
        if "t" in results:
            results["t"] = results["t"][t_APs]
        for key in keys_to_sample:
            results[key] = results[key][np.ix_(I_x, t_APs)]
        ###############################
        ## remove non nevessary data ##
        ###############################
        list_keys =  {
        "ID",
        "model",
        "x_rec",
        "rec",
        "Nseg_per_sec",
        "axon_path_type",
        "t_sim",
        "myelinated",
        "intra_stim_starts",
        "intra_stim_positions",
        "recorder"
        }

        list_keys.update(keys_to_keep)
        list_keys.update(keys_to_sample)
        if results.ID==0:
            list_keys.update({"t"})
        results.remove_key(keys_to_keep=list_keys, keys_to_remove=keys_to_remove)
    return results

def sample_g_mem(results:axon_results, t_start_rec:float=0, t_stop_rec:float=-1, sample_dt:None|float=None, x_bounds:None|float|tuple[float]=None)->axon_results:
    """
    Undersample the membrane coductivity (``results["g_mem"]``) key and remove most of the `axon_results` keys to alliviate RAM usage.

    Note
    ----
    Only the following keys kept by the fuction
     - `g_mem`
     - `x_rec`
     - `rec`
     - `Nseg_per_sec`
     - `axon_path_type`
     - `t_sim`

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    t_start_rec : float, optional
        Lower time at whitch `g_mem` should be stored, by default 0
    t_stop_rec : float, optional
        Upper time at whitch `g_mem` should be stored, by default -1
    sample_dt : None | float, optional
        Time sample rate at which `g_mem` should be stored if None simulation dt is kept, by default None
    x_bounds : None | tuple[float], optional
        x-positions where to store `g_mem`, possible option:
         - float: The values of `g_mem` are only stored at the nearest position in `x_rec`.
         - tupple: `g_mem` values are stored for all positions included between the two boundaries.
         - None (default): `g_mem` values are stored for all positions.

    Warning
    -------
    ``sample_dt`` should be at multiple of the simulation ``dt`` to allow a correct undersampling. 
    If the not ``sample_dt`` choosen will be the closer multiple of ``dt``.

    Returns
    -------
    axon_results
        updated results.
    """
    if not results["record_g_mem"]:
        rise_error("gmem not recorded nothing will be done")
    else:
        if t_stop_rec < 0:
            t_stop_rec=results.t_sim

        if sample_dt is None:
            sample_dt=results.dt

        if x_bounds is None:
            x_bounds=(0,results.L)

        if np.iterable(x_bounds):
            I_x = np.argwhere((results["x_rec"]>x_bounds[0])&(results["x_rec"]<x_bounds[1]))[:,0]
        else:
            x_bounds = [x_bounds]
            I_x = np.array([np.argmin(abs(results["x_rec"]-x_bounds[0]))])

        N_x = len(I_x)
        i_t_min = np.argwhere(results["t"]>t_start_rec)[0][0]
        i_t_max = np.argwhere(results["t"]<t_stop_rec)[-1][0]

        t_APs = [k for k in range(i_t_min,i_t_max)]
        t_APs = t_APs[::int(sample_dt/results.dt)]
        N_t = len(t_APs)


        # Under sampling to reduce memory consumption
        results["x_rec"] = results["x_rec"][I_x] - x_bounds[0]
        results["t"] = results["t"][t_APs]
        results["g_mem"] = results["g_mem"][np.ix_(I_x, t_APs)]

        ###############################
        ## remove non nevessary data ##
        ###############################
        list_keys = {
        "g_mem",
        "x_rec",
        "rec",
        "Nseg_per_sec",
        "axon_path_type",
        "t_sim",
        }
        if results.ID==0:
            list_keys.update({"t"})
        results.remove_key(keys_to_keep=list_keys)
    return results

def vmem_plot(results:axon_results, freq:float=None, save:bool=False, fdir:str=""):
    """
    Plot and save the membrane potential along each axon of the fascicle

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    freq : float, optional
            Filter vmem at the specified frequency, by default None
    save : bool, optional
        if true, the block status is saved as a line in a `.csv` file, by default False
    fdir : str, optional
        Path where the conductivity should be saved, the saving file will be `fdir+"block_summary.csv"`, by default ""

    Returns
    -------
    axon_results
        updated results.
    """
    vm_key = "V_mem"
    if (freq is not None):
        results.filter_freq("V_mem",freq , Q= 2)
        vm_key += "_filtered"
    
    fig, ax = plt.subplots()
    results.plot_x_t(ax, key=vm_key)
    if results.myelinated:
        title = f"Myelinated Axon: {np.round(results.diameter,2)} µm in diameter"
    else:       #colormap for unmyelinated fibers as plot_x_t are unreadable
        title = f"Unmyelinated Axon: {np.round(results.diameter,2)} µm in diameter"
    
    #print(fdir)
    ax.set_title(title)
    if save:
        fig.tight_layout()
        fig_name = fdir + "/Activity_axon_" + str(results.ID) + ".png"
        fig.savefig(fig_name)
        plt.close(fig)
    return results

def raster_plot(results:axon_results, freq:float=None, save:bool=False, fdir:str=""):
    """
    Plot and save the raster plot along each axon of the fascicle.

    Parameters
    ----------
    results : axon_results
        results of the axon simulation.
    freq : float, optional
            Filter vmem at the specified frequency, by default None
    save : bool, optional
        if true, the block status is saved as a line in a `.csv` file, by default False
    fdir : str, optional
        Path where the conductivity should be saved, the saving file will be `fdir+block_summary.csv`, by default ""

    Returns
    -------
    axon_results
        updated results.
    """

    vm_key = "V_mem"
    if (freq is not None):
        results.filter_freq("V_mem",freq , Q= 2)
        vm_key += "_filtered"
    fig, ax = plt.subplots()

    results.raster_plot(ax, key=vm_key)
    if results.myelinated:
        title = f"Myelinated Axon: {np.round(results.diameter,2)} µm in diameter"
    else:
        title = f"Unyelinated Axon: {np.round(results.diameter,2)} µm in diameter"

    #print(fdir)
    ax.set_title(title)
    if save:
        fig.tight_layout()
        fig_name = fdir + "/Rasterplot_axon_" + str(results.ID) + ".png"
        fig.savefig(fig_name)
        plt.close(fig)
    return results


"""
NRV-:class:`.axon_results` handling.
"""

import numpy as np
from numpy.typing import NDArray
from itertools import combinations
import matplotlib.pyplot as plt
from copy import deepcopy
from scipy.signal import find_peaks

from ...fmod.FEM.fenics_utils._f_materials import f_material, mat_from_interp
from ...backend._NRV_Results import sim_results, abstractmethod
from ...backend._file_handler import json_dump
from ...backend._log_interface import rise_warning, rise_error
from ...utils._units import to_nrv_unit, from_nrv_unit, convert, nm
from ...utils._misc import (
    distance_point2line,
    membrane_capacitance_from_model,
    compute_complex_admitance,
    nearest_greater_idx,
)


def max_spike_position(blocked_spike_positionlist, position_max, spike_begin="down"):
    rise_warning("DeprecationWarning: ", "max_spike_position is obsolete")
    if spike_begin == "down":
        while blocked_spike_positionlist[
            position_max + 1
        ] >= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max
    else:
        while blocked_spike_positionlist[
            position_max + 1
        ] <= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max


def AP_detection(
    Voltage, t, x, list_to_parse, thr, dt, t_start, t_stop, t_refractory, t_min_spike
):
    """
    Internal use only, detects an AP
    """

    rise_warning("DeprecationWarning: ", "AP_detection is obsolete")
    raster_position = []
    raster_x_position = []
    raster_time_index = []
    raster_time = []
    # parsing to find spikes

    for i in list_to_parse:
        t_last_spike = t_start - t_refractory
        for j in range(int(t_start * (1 / dt)), t_stop):
            if (
                Voltage[i][j] <= thr
                and Voltage[i][j + 1] >= thr
                and Voltage[i][min((j + int(t_min_spike * (1 / dt))), t_stop)] >= thr
                and (j * dt - t_last_spike) > t_refractory
            ):  # 1st line: threshold crossing, 2nd: minimum time above threshold,3rd: refractory period
                # there was a spike, get time and position
                raster_position.append(i)
                raster_x_position.append(x[i])
                raster_time_index.append(j)
                raster_time.append(t[j])
                # memorize the time in ms, to evaluate refractory period
                t_last_spike = j * dt
    # return results
    return (
        np.asarray(raster_position),
        np.asarray(raster_x_position),
        np.asarray(raster_time_index),
        np.asarray(raster_time),
    )


# @jit(nopython=True, fastmath=True)
def count_spike(onset_position):
    """
    spike counting, just in time compiled. For internal use only.
    """

    rise_warning("DeprecationWarning: ", "AP_detection is obsolete")
    if len(onset_position) == 0:
        spike_number = 0
        return 0
    else:
        spike_number = 1
        for i in range(len(onset_position) - 1):
            if onset_position[i] == min(onset_position):
                if onset_position[i] == onset_position[i + 1]:
                    spike_number = spike_number + 1
    return spike_number


def rasterize(
    y_data: np.array,
    x_data: np.array,
    t_data: np.array,
    threshold: np.float64,
    t_min_AP=0.1,
    t_refractory=0.5,
    t_start: np.float64 = None,
    t_stop: np.float64 = None,
) -> np.array:
    """
    Rasterize a membrane potential (or filtered or any quantity processed from membrane voltage), with AP detection.

    Parameters
    ----------
    y_data : np.array
        Data to rasterize (membrane potential, or anything else)
    x_data : np.array
        x coordinates of y_data
    t_data : np.array
        t coordinates of y_data
    threshold : np.float64
        threshold for AP detection
    t_min_AP : float, optional
        Mininum AP duration, in ms, by default 0.1ms
    t_refractory : int, optional
        inter-AP duration, in ms. Default is 2ms, by default 2
    t_start : np.float64, optional
        start time for rasterize, in ms. Default is 0ms, by default None
    t_stop : np.float64, optional
        stop time for rasterize, in ms. Default is simulation time., by default None
    """

    if t_start is not None:
        t_start_idx = np.where(t_data >= t_start)[0][0]
    else:
        t_start_idx = 0

    if t_stop is not None:
        t_stop_idx = np.where(t_data <= t_stop)[0][0]
    else:
        t_stop_idx = -1

    t_idx_AP = np.array([], np.int32)
    x_idx_AP = np.array([], np.int32)

    dt = t_data[1]-t_data[0]
    dis = int(t_refractory/dt)
    width = int(t_min_AP/dt)
    #width = None
    height = [-20,70]     #min/max heigh of AP

    for x_val, v in enumerate(y_data):
        data = v[t_start_idx:t_stop_idx]
        peak_idxs, _ = find_peaks(data, height=height,distance=dis, width=width)
        t_idx_AP = np.concatenate([t_idx_AP, peak_idxs])
        x_idx_AP = np.concatenate([x_idx_AP, np.ones(np.size(peak_idxs)) * x_val])
        
    t_idx_AP = np.int32(t_idx_AP)
    x_idx_AP = np.int32(x_idx_AP)

    if len(t_data) :
        return (x_idx_AP, x_data[x_idx_AP], t_idx_AP, t_data[t_idx_AP])
    else:
        return (np.array([]), np.array([]), np.array([]), np.array([]))


def get_first_AP(
    x_APs: np.array,
    x_idx_APs: np.array,
    t_APs: np.array,
    t_idx_APs: np.array,
    t_refract: np.float64 = 0.5,
) -> NDArray | NDArray | NDArray:
    """
    Extract time and position of the first rasterized AP

    Parameters
    ----------
    x_APs : np.array
        x position of the rasterized APs
    x_idx_APs : np.array
        x position index of the rasterized APs
    t_APs : np.array
        t position of the rasterized APs
    t_idx_APs : np.array
        t position index of the rasterized APs
    t_refract: np.float64
        fiber refractory time, in ms, default is .5ms

    Returns
    -------
    np.array
        x positions of the first AP
    np.array
        x positions index of the first AP
    np.array
        t positions of the first AP
    np.array
        t positions index of the first AP
    np.array
        index position of the AP
    """

    if len(t_APs) == 1:
        return (np.array(x_APs), [0], np.array(t_APs), [0], [0])

    if len(t_APs) < 1:
        return (np.array([]), np.array([]), np.array([]), np.array([]), np.array([]))

    ordered_t_idx = np.argsort(t_APs)  # sort time indexes

    x_ordered = x_APs[ordered_t_idx]  # sort x positions accordingly
    t_ordered = t_APs[ordered_t_idx]  # sort t positions accordingly
    x_idx_ordered = x_idx_APs[ordered_t_idx]  # sort x positions idx accordingly
    t_idx_ordered = t_idx_APs[ordered_t_idx]  # sort t positions idx accordingly

    x_0 = x_ordered[0]  # get x-position of the first AP in time
    idx_unique = np.sort(np.unique(x_ordered, return_index=True, axis=0)[1])
    t_min = t_refract
    x_ordered_unique = x_ordered[idx_unique]
    x_ordered_unique.sort()
    if len(x_ordered_unique) < 2:
        return (np.array([]), np.array([]), np.array([]), np.array([]), np.array([]))
    
    x_min = np.min(np.abs(np.diff(x_ordered_unique))) * 5


    msk_unique = np.arange(len(t_ordered) - 1)
    msk_unique = [i for i in msk_unique if i not in idx_unique]

    # get upward APs
    idx_up = []
    idx_up.append(0)  # add the first one
    mask = np.full(x_ordered.shape, False)
    mask[msk_unique] = True
    mask[np.where(x_ordered < x_0)[0]] = True  # mask downward APs

    mask[0] = True  # mask first AP
    x_ordered_cpy = x_ordered.copy()  # copy data
    x_ordered_cpy[mask] = np.nan  # add nan where mask is true
    keep_going = True

    while keep_going:
        if (np.isnan(x_ordered_cpy).sum()) == len(
            x_ordered_cpy
        ):  # check if array is full of nan
            keep_going = False
        else:
            idx = np.nanargmin((x_ordered_cpy - x_0))  # get closest value
            x_ordered_cpy[idx] = np.nan  # mask it

            if (
                t_ordered[idx] - t_ordered[idx_up[-1]] >= 0
            ):  # if time val is greater than previous then we reached of end of the AP
                if (
                    t_ordered[idx] - t_ordered[idx_up[-1]] < t_min
                ):  # if time than the AP refractory period then it doesn't belong to the AP
                    if (
                        np.abs(x_ordered[idx] - x_ordered[idx_up[-1]]) < x_min
                    ):  # if it "jumps" to far then it doesn't belong to the AP
                        idx_up.append(idx)

    # get downward APs
    mask = np.full(x_ordered.shape, False)
    idx_neg = np.where(x_ordered > x_0)[0]
    mask[idx_neg] = True
    mask[0] = True
    mask[msk_unique] = True
    x_ordered_cpy = x_ordered.copy()
    x_ordered_cpy[mask] = np.nan
    idx_down = []
    idx_down.append(0)
    keep_going = True
    while keep_going:
        if (np.isnan(x_ordered_cpy).sum()) == len(x_ordered_cpy):
            keep_going = False
        else:
            idx = np.nanargmin(-(x_ordered_cpy - x_0))
            x_ordered_cpy[idx] = np.nan
            if (t_ordered[idx] - t_ordered[idx_down[-1]]) >= 0:
                if (t_ordered[idx] - t_ordered[idx_down[-1]]) <= t_min:
                    if np.abs(x_ordered[idx] - x_ordered[idx_down[-1]]) < x_min:
                        idx_down.append(idx)

    idx_APs = idx_up + idx_down
    AP_indexes = np.int32(idx_APs)
    AP_indexes = np.unique(idx_APs)

    return (
        x_ordered[AP_indexes],
        x_idx_ordered[AP_indexes],
        t_ordered[AP_indexes],
        t_idx_ordered[AP_indexes],
        ordered_t_idx[AP_indexes],
    )


class axon_results(sim_results):
    """"""

    def __init__(self, context=None):
        super().__init__(context)

    def generate_axon(self):
        """
        generate from the results new version of the simulated axon. Axon generated has the same parameters but is not simulated yet.

        Note
        ----
        This function is not defined in ``axon_results`` but in ``unmyelinated_results`` and ``myelinated_results``.
        """
        pass

    def is_recruited(self, vm_key: str = "V_mem", t_start: float = None) -> bool:
        """
        Return True if an AP is detected, else False.

        Returns
        -------
        is_recruited                : bool
            Return True if an AP is detected, else False.
        """
        if not "recruited" in self:
            n_aps = self.count_APs(vm_key) != 0
            if n_aps :
                if t_start is None: 
                    self["recruited"] = n_aps
                else:
                    _,t_starts = self.get_start_APs(vm_key= vm_key)
                    #print(t_start)
                    #print(t_starts)
                    #print(t_starts[t_starts>=t_start])
                    #print(len(t_starts[t_starts>=t_start])>0)
                    self["recruited"] = len(t_starts[t_starts>=t_start])>0
            else:
                self["recruited"] = n_aps
        return self["recruited"]

    def is_blocked(
        self, AP_start: float, freq: float = None, t_refractory: float = 1
    ) -> bool | None:
        """
        check if the axon is blocked or not.

        Warning
        -------
        A test pulse must be added to generate an AP propagating through the axon. This test AP is used to
        validate or not the neural conduction of the axon.

        Parameters
        ----------
        AP_start : float
            timestamp of the test pulse start, in ms.
        freq : float, optional
            Frequency of the stimulation, for KES block, by default None
        t_refractory : float, optional
            Axon refractory period, by default 1

        Returns
        -------
        bool
            True is the axon is blocked (the test AP doesn't propagate through), else false.
        """
        if not "blocked" in self:
            vm_key = "V_mem"
            if freq is not None:
                self.filter_freq("V_mem", freq, Q=2)
                vm_key += "_filtered"
            self.rasterize(vm_key=vm_key,clear_artifacts = False)   #artifacts clearing produces strange results :(
            _, t_starts = self.get_start_APs(vm_key)
            n_APs = len(t_starts[t_starts >= AP_start])

            t_AP_start = None
            if n_APs > 0:
                test_AP_idx = nearest_greater_idx(
                    t_starts, AP_start
                )  # get idx of test AP in the APs list
                x_APs, _, t_APs, _ = self.split_APs(vm_key)
                x_AP_test = x_APs[test_AP_idx]
                t_AP_test = t_APs[test_AP_idx]
                _, t_AP_start = self.get_start_AP(x_AP_test, t_AP_test)

            delta = 1
            
            if n_APs == 0 or (
                t_AP_start is not None and t_AP_start>= AP_start + delta
            ):  
                rise_warning(
                    f"is_blocked in Axon {self.ID}: No AP detected after the test pulse. Make sure the stimulus is strong enough, attached to the axon or not colliding with onset response. \n ... Fiber considered as blocked."
                )
                self["blocked"] = False
                return self["blocked"]
            if n_APs > 2:
                rise_warning(
                    f"is_blocked in Axon {self.ID}: More than two APs are detected after the test pulse, likely due to onset response. This can cause erroneous block state estimation. \n ... Consider increasing axon's spatial discretization."
                )
                self["blocked"] = False
                return self["blocked"]

            if self.has_AP_reached_end(
                x_AP_test, t_AP_test
            ):  # test AP has reached the end --> is not blocked
                self["blocked"] = False
                return self["blocked"]
            else:
                if not self.is_AP_in_timeframe(x_AP_test, t_AP_test):  #
                    rise_warning(
                        f"is_blocked in Axon {self.ID}: Test AP didn't not reach axon ends within the simulation timeframe. Consider increasing simulation time or start the test stimulus earlier. "
                    )
                    return None
                _, _, coll_list = self.get_collision_pts(vm_key)
                if coll_list[test_AP_idx]:
                    rise_warning(
                        f"is_blocked in Axon {self.ID}: Test AP is colliding with an other AP, probably onset response. Consider increasing the duration between start of block stimulus and test stimulus."
                    )
                    return None

                t_AP_test_max = np.max(t_AP_test)  # test AP last time position:

                if len(t_starts[t_starts < AP_start]):
                    if (
                        np.min(np.abs(t_starts[t_starts < AP_start] - t_AP_test_max))
                        < t_refractory
                    ):  # check if test AP is during refractory period
                        rise_warning(
                            f"is_blocked in Axon{self.ID}: Test AP occures during axon's refractory period, probably due to onset response. Consider increasing the duration between start of block stimulus and test stimulus."
                        )
                        return None

                if len(t_starts[t_starts > t_AP_start]):  # AP jump
                    if (
                        np.min(np.abs(t_starts[t_starts > AP_start] - t_AP_test_max))
                        < t_refractory
                    ):
                        # todo: make sure that the jumping AP is reaching the over side of the axon
                        self["blocked"] = False
                        return self["blocked"]
                self["blocked"] = True
                return self["blocked"]
        return self["blocked"]

    def split_APs(
        self, vm_key: str = "V_mem"
    ) -> list[np.array] | list[np.array] | list[np.array] | list[np.array]:
        """
        Detects individual Action potential in vm_key and split them in lists

        Parameters
        ----------
        vm_key : str, optional
             vmembrane key, by default "V_mem"

        Returns
        -------
        list[np.array]
            list of the x_pos of each AP
        list[np.array]
            list of the x index of each AP
        list[np.array]
            list of the time of each AP
        list[np.array]
            list of the time index of each AP
        """
        # if not vm_key + "_raster_x_position" in self:
        if not vm_key + "_raster_x_position" in self:
            self.rasterize(vm_key=vm_key, clear_artifacts=False)
        x_APs = []
        x_idx_APs = []
        t_APs = []
        t_idx_APs = []
        x_raster = self[vm_key + "_raster_x_position"].copy()
        t_raster = self[vm_key + "_raster_time"].copy()
        x_idx_raster = self[vm_key + "_raster_position"].copy()
        t_idx_raster = self[vm_key + "_raster_time_index"].copy()

        min_size = self._get_min_AP_length()
        #plt.figure()
        if len(x_raster) > min_size:
            while len(x_raster) > min_size:
                x_AP, x_idx_AP, t_AP, t_idx_AP, AP_idx = get_first_AP(
                    x_raster, x_idx_raster, t_raster, t_idx_raster
                )
                if len(x_AP) == 0:
                    AP_len = 0
                else:
                    AP_len = np.max(
                        [
                            self.get_AP_upward_len(x_AP, t_AP),
                            self.get_AP_downward_len(x_AP, t_AP),
                        ]
                    )
                #print(f"upward: {self.get_AP_upward_len(x_AP, t_AP)} - downward: {self.get_AP_downward_len(x_AP, t_AP)}")
                #plt.scatter(t_AP,x_AP)
                if AP_len >= min_size:
                    x_APs.append(x_AP)
                    t_APs.append(t_AP)
                    x_idx_APs.append(x_idx_AP)
                    t_idx_APs.append(t_idx_AP)
                if len(AP_idx):
                    x_raster = np.delete(x_raster, AP_idx)
                    t_raster = np.delete(t_raster, AP_idx)
                    x_idx_raster = np.delete(x_idx_raster, AP_idx)
                    t_idx_raster = np.delete(t_idx_raster, AP_idx)
                else:
                    x_raster = np.array([])
                    t_raster = np.array([])
                    x_idx_raster = np.array([])
                    t_idx_raster = np.array([])

        #plt.show()
        #exit()
        return (x_APs, x_idx_APs, t_APs, t_idx_APs)

    def count_APs(self, vm_key: str = "V_mem") -> int:  # **kwargs
        """
        Count number of APs detected in vm_key

        Parameters
        ----------
        vm_key : str, optional
            Vmembrane key, by default "V_mem"

        Returns
        -------
        int
            number of APs detected
        """
        x, _, _, _ = self.split_APs(vm_key)
        return len(x)

    def _get_min_AP_length(self, min_size: float = 0.1) -> int:
        """
        Returns the minimum propation length of an AP, set to 10% of the axon length by default

        Parameters
        ----------
        min_size : float, optional
            minimum AP length, with respect to the fiber length, by default .1

        Returns
        -------
        int
            return the minimum AP length pts
        """
        if self.myelinated == True and self.rec == "all":
            n_x = len(self.node_index)
        else:
            n_x = len(self.x_rec)
        return int(n_x * min_size)

    def get_start_AP(self, x_AP: np.array, t_AP: np.array) -> float | float:
        """
        Returns the x,t start values of the AP

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        float
            x start value of the AP
        float
            t start value of the AP
        """
        start_idx = t_AP.argmin()
        return (x_AP[start_idx], t_AP[start_idx])

    def get_start_APs(self, vm_key: str = "V_mem") -> NDArray | NDArray:
        """
        Return list of (x,t) start positions of each detected APs

        Parameters
        ----------
        vm_key : str, optional
            Vmembrane key, by default "V_mem", by default "V_mem"

        Returns
        -------
        Returns
        -------
        NDArray
            Array of the start x-position of each APs
        NDArray
            Array of the start t-position of each APs
        """
        x_APs, _, t_APs, _ = self.split_APs(vm_key)
        x_AP_start = []
        t_AP_start = []
        for x_AP, t_AP in zip(x_APs, t_APs):
            x_AP_s, t_AP_s = self.get_start_AP(x_AP, t_AP)
            x_AP_start.append(x_AP_s)
            t_AP_start.append(t_AP_s)
        return (np.array(x_AP_start), np.array(t_AP_start))

    def get_xmax_AP(self, x_AP: np.array, t_AP: np.array) -> float | float:
        """
        Returns the x,t values the maximum AP x-position

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        float
            x value of the maximum AP x-position
        float
            t value of the maximum AP x-position
        """
        max_idx = x_AP.argmax()
        return (x_AP[max_idx], t_AP[max_idx])

    def get_xmin_AP(self, x_AP: np.array, t_AP: np.array) -> float | float:
        """
        Returns the x,t values the minimum AP x-position

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        float
            x value of the minimum AP x-position
        float
            t value of the minimum AP x-position
        """
        min_idx = x_AP.argmin()
        return (x_AP[min_idx], t_AP[min_idx])

    def get_AP_upward_len(self, x_AP: np.array, t_AP: np.array) -> int:
        """
        Returns the length (number of x pts) of the AP in the upward direction

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        int
            number of AP x-pos points in the upward direction
        """
        x_start, _ = self.get_start_AP(x_AP, t_AP)
        x_max, _ = self.get_xmax_AP(x_AP, t_AP)
        len_upward = x_AP[((x_AP >= x_start) & (x_AP <= x_max))]
        return len(len_upward)

    def get_AP_downward_len(self, x_AP: np.array, t_AP: np.array) -> int:
        """
        Returns the length (number of x pts) of the AP in the downward direction

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        int
            number of AP x-pos points in the downward direction
        """
        x_start, _ = self.get_start_AP(x_AP, t_AP)
        x_min, _ = self.get_xmin_AP(x_AP, t_AP)
        len_downward = x_AP[((x_AP <= x_start) & (x_AP >= x_min))]
        return len(len_downward)

    def has_AP_reached_end(self, x_AP: np.array, t_AP: np.array) -> bool:
        """
        Return true if the AP has reached the end of the axon in both direction, else false

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        bool
            is true if the AP has reached the end of the axon in both direction, else false
        """
        x = self.get_axon_xrec()
        x_max = np.max(x)
        x_min = np.min(x)
        x_max_AP, _ = self.get_xmax_AP(x_AP, t_AP)
        x_min_AP, _ = self.get_xmin_AP(x_AP, t_AP)
        return x_max == x_max_AP and x_min_AP == x_min

    def APs_reached_end(self, vm_key: str = "V_mem") -> bool:
        """
        Return true if ALL APs have reached the both ends of the axon

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        bool
            True if ALL APs have reached the both ends of the axon, else False
        """

        x_APs, _, t_APs, _ = self.split_APs(vm_key=vm_key)
        if len(x_APs):
            for x_AP, t_AP in zip(x_APs, t_APs):
                if not self.has_AP_reached_end(x_AP, t_AP):
                    return False
        return True

    def is_AP_in_timeframe(self, x_AP: np.array, t_AP: np.array) -> bool:
        """
        Return true if AP has reached it's last position (in both direction) within the simulation timeframe, else False

        Parameters
        ----------
        x_AP : np.array
            x positions of the AP
        t_AP : np.array
            time values of the AP

        Returns
        -------
        bool
            true if AP has reached its last position (in both direction) within the simulation timeframe, else False
        """
        if self.has_AP_reached_end(x_AP, t_AP):
            return True
        else:
            up_len = self.get_AP_upward_len(x_AP, t_AP)
            down_len = self.get_AP_downward_len(x_AP, t_AP)
            _, t_start = self.get_start_AP(x_AP, t_AP)
            if up_len > down_len:
                _, t_stop = self.get_xmax_AP(x_AP, t_AP)
                dt_AP = (t_stop - t_start) / up_len
            else:
                _, t_stop = self.get_xmin_AP(x_AP, t_AP)
                dt_AP = (t_stop - t_start) / down_len

            if dt_AP < 0.5:
                dt_AP = 0.5
            _, t_xmax_AP = self.get_xmax_AP(x_AP, t_AP)
            _, t_xmin_AP = self.get_xmin_AP(x_AP, t_AP)
            tmax = np.max(self.t)
            if (t_xmax_AP + dt_AP > tmax) or (t_xmin_AP + dt_AP > 10 * tmax):
                return False
            return True

    def APs_in_timeframe(self, vm_key: str = "V_mem") -> bool:
        """
        Return true if ALL APs have reached their last position (in both direction) within the simulation timeframe, else False

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        bool
            True if ALL APs have reached their last position (in both direction) within the simulation timeframe, else False
        """
        x_APs, _, t_APs, _ = self.split_APs(vm_key=vm_key)
        if len(x_APs):
            for x_AP, t_AP in zip(x_APs, t_APs):
                if not self.is_AP_in_timeframe(x_AP, t_AP):
                    return False
        return True

    def get_collision_pts(self, vm_key: str = "V_mem") -> list[float] | list[float]:
        """
        Returns three lists: - two containing the x,t coordinates of the interAPs colliding coordinates
                             - one boolean list where true means the AP is colliding, else false

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        list[float]
            x coordinates of the interAPs collision point
        list[float]
            t coordinates of the interAPs collision point
        list[bool]
            list of colliding APs: true when colliding else false
        """
        x_APs, _, t_APs, _ = self.split_APs(vm_key=vm_key)
        x_APs_coll = []
        t_APs_coll = []
        idx_coll = []
        colliding_list = [False for i in range(len(x_APs))]
        idx_APs = np.arange(len(x_APs))

        for x_AP, t_AP, idx_AP in zip(x_APs, t_APs, idx_APs):
            if not self.has_AP_reached_end(x_AP, t_AP):
                x_APs_coll.append(x_AP)
                t_APs_coll.append(t_AP)
                idx_coll.append(idx_AP)
        if len(x_APs_coll) <= 1:  # need at least 2 APs to get a collision
            return ([], [], colliding_list)
        ids = np.arange(len(x_APs_coll))
        comb_list = list(combinations(ids, 2))
        x_coll_l = []
        t_coll_l = []
        for idx_comb in comb_list:
            x_inter, t_inter = self.get_interAPs_collision(
                x_APs_coll[idx_comb[0]],
                t_APs_coll[idx_comb[0]],
                x_APs_coll[idx_comb[1]],
                t_APs_coll[idx_comb[1]],
            )
            if (x_inter and t_inter) > 0:
                x_coll_l.append(x_inter)
                t_coll_l.append(t_inter)
                colliding_list[idx_coll[idx_comb[1]]] = True
                colliding_list[idx_coll[idx_comb[0]]] = True

        return (x_coll_l, t_coll_l, colliding_list)

    def detect_AP_collisions(self, vm_key: str = "V_mem") -> bool:
        """
        Return True is at least one collision between two APs is detected, else False

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        bool
            True if an interAP collision is detected, else False
        """
        _, t_coll, _ = self.get_collision_pts(vm_key)
        if len(t_coll):
            return True
        return False

    def get_interAPs_collision(
        self, x_AP1: NDArray, t_AP1: NDArray, x_AP2: NDArray, t_AP2: NDArray
    ) -> float | float:
        """
        Returns the estimated collision coordinates of two APs. Returns (0,0) if the collision is out of scope.

        Parameters
        ----------
        x_AP1 : NDArray
            x positions of the first AP
        t_AP1 : NDArray
            t positions of the first AP
        x_AP2 : NDArray
            x positions of the seconc AP
        t_AP2 : NDArray
            t positions of the second AP
        Returns
        -------
        float
            x position of the intersection (collision) point. Is 0 is the position is out of scope
        float
            t position of the intersection (collision) point. Is 0 is the position is out of scope
        """
        poly_up_1, poly_down_1 = self.linfit_AP(x_AP1, t_AP1)
        poly_up_2, poly_down_2 = self.linfit_AP(x_AP2, t_AP2)
        x_inter1, t_inter1 = self.get_1dpoly_intersec(poly_up_1, poly_down_2)
        x_inter2, t_inter2 = self.get_1dpoly_intersec(poly_up_2, poly_down_1)

        x1_max, t1_max = self.get_xmax_AP(x_AP1, t_AP1)
        x1_min, t1_min = self.get_xmin_AP(x_AP1, t_AP1)

        x2_max, t2_max = self.get_xmax_AP(x_AP2, t_AP2)
        x2_min, t2_min = self.get_xmin_AP(x_AP2, t_AP2)

        if self._pts_in_sim_range(x_inter1, t_inter1) and self._pts_in_sim_range(
            x_inter2, t_inter2
        ):
            if t_inter2 > t_inter1:
                if self._AP_pt_in_range(
                    x1_max, t1_max, x_inter2, t_inter2
                ) and self._AP_pt_in_range(x2_min, t2_min, x_inter2, t_inter2):
                    if t_inter2 > (t1_max and t2_min):
                        return x_inter2, t_inter2
                if self._AP_pt_in_range(
                    x1_min, t1_min, x_inter2, t_inter2
                ) and self._AP_pt_in_range(x2_max, t2_max, x_inter2, t_inter2):
                    if t_inter2 > (t1_min and t2_max):
                        return x_inter2, t_inter2
            else:
                if self._AP_pt_in_range(
                    x1_max, t1_max, x_inter1, t_inter1
                ) and self._AP_pt_in_range(x2_min, t2_min, x_inter1, t_inter1):
                    if t_inter1 > (t1_max and t2_min):
                        return x_inter1, t_inter1
                if self._AP_pt_in_range(
                    x1_min, t1_min, x_inter1, t_inter1
                ) and self._AP_pt_in_range(x2_max, t2_max, x_inter1, t_inter1):
                    if t_inter1 > (t1_min and t2_max):
                        return x_inter1, t_inter1
        return 0, 0

    def _pts_in_sim_range(self, x: float, t: float) -> bool:
        """
        Returns true if the (x,t) point is within the simulation range

        Parameters
        ----------
        x : float
            x coordinate of the point
        t : float
            t coordinate of the point

        Returns
        -------
        bool
            true if the (x,t) point is within the simulation range, else false
        """
        x_max = np.max(self.get_axon_xrec())
        t_max = np.max(self.t)
        return (x < x_max) and (x > 0) and (t > 0) and (t < t_max)

    def _AP_pt_in_range(
        self, x1: float, t1: float, x2: float, t2: float, tol: float = 0.12
    ) -> bool:
        """
        Return true is the point (x1,t1) and (x2,t2) are within a range, set by tol

        Parameters
        ----------
        x1 : float
            x coordinate of the (x1,t1) point
        t1 : float
            t coordinate of the (x1,t1) point
        x2 : float
            x coordinate of the (x2,t2) point
        t2 : float
            t coordinate of the (x2,t2) point
        tol : float, optional
            range tolerance, expressed as a fraction of the (x1,t1) pts. by default 0.1

        Returns
        -------
        bool
            True if the two points are within the range
        """
        return np.abs(x1 - x2) <= tol * x1 and np.abs(t1 - t2) <= tol * t1

    def get_1dpoly_intersec(self, poly1: np.poly1d, poly2: np.poly1d) -> float | float:
        """
        Returns the intersection points of two 1D poly

        Parameters
        ----------
        poly1 : np.poly1d
            First 1D poly
        poly2 : np.poly1d
            Second 1D poly

        Returns
        -------
        float
            y intersection point
        float
            x intersection point
        """
        x_inter = (poly2.c[1] - poly1.c[1]) / (poly1.c[0] - poly2.c[0])
        return (poly1(x_inter), x_inter)

    def linfit_AP(self, x_AP: NDArray, t_AP: NDArray) -> np.poly1d | np.poly1d:
        """
        Fit the AP with two 1D polynomial equations: one for the upward propagation, one for the downward propagation

        Parameters
        ----------
        x_AP : NDArray
            x positions of the AP
        t_AP : NDArray
            t positions of the AP

        Returns
        -------
        np.poly1d
            AP upward direction 1D poly fit
        np.poly1d
            AP downward direction 1D poly fit
        """
        x_max, _ = self.get_xmax_AP(x_AP, t_AP)
        x_min, _ = self.get_xmin_AP(x_AP, t_AP)
        x_start, _ = self.get_start_AP(x_AP, t_AP)

        # fit upward AP
        upward_idx = np.argwhere((x_AP >= x_start) & (x_AP <= x_max))
        upward_idx = [x[0] for x in upward_idx]
        x_AP_up = x_AP[upward_idx]
        t_AP_up = t_AP[upward_idx]
        idx_sorted = np.argsort(t_AP_up)
        t_AP_up = t_AP_up[idx_sorted]
        x_AP_up = x_AP_up[idx_sorted]
        p_up = np.poly1d(np.polyfit(t_AP_up, x_AP_up, 1))

        # fit downward AP
        downward_idx = np.argwhere((x_AP <= x_start) & (x_AP >= x_min))
        downward_idx = [x[0] for x in downward_idx]
        x_AP_down = x_AP[downward_idx]
        t_AP_down = t_AP[downward_idx]
        idx_sorted = np.argsort(t_AP_down)
        t_AP_down = t_AP_down[idx_sorted]
        x_AP_down = x_AP_down[idx_sorted]
        p_down = np.poly1d(np.polyfit(t_AP_down, x_AP_down, 1))
        return (p_up, p_down)

    def get_axon_xrec(self) -> NDArray:
        """
        Return the  x-position recorded array

        Returns
        -------
        NDArray
            x-position recorded array
        """

        if self["myelinated"] == True:
            if self["rec"] == "All":
                x_idx = self["node_index"]
                x = self["x"][x_idx]
            else:
                x = self["x_rec"]
        else:
            x = self["x_rec"]
        return x

    def get_avg_AP_speed(self, vm_key: str = "V_mem") -> float:
        """
        Returns the averaged propagtion speed of each APs, in m/s.

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        float
            averaged propagation speed of each APs, in m/s. Returns 0.0 if no AP is detected for speed evaluation.
        """
        ap_speed = self.getAPspeed(vm_key)
        if len(ap_speed):
            return np.round(np.mean(ap_speed), 3)
        else:
            rise_warning("No AP detected, can't evaluate AP propagation speed.")
            return 0.0

    def getAPspeed(self, vm_key: str = "V_mem") -> list[float]:
        """
        Return the propagtion speed of each APs, in m/s

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        Returns
        -------
        list[float]
            list of propagation speed of each APs, in m/s
        """
        x_APs, _, t_APs, _ = self.split_APs(vm_key=vm_key)
        APspeed = []
        if len(x_APs):
            for x_AP, t_AP in zip(x_APs, t_APs):
                APspeed.append(np.round(self._APspeed(x_AP, t_AP), 3))
        if len(APspeed):
            return APspeed
        else:
            rise_warning("No AP detected, can't evaluate AP propagation speed.")
            return []

    def _APspeed(self, x_AP: NDArray, t_AP: NDArray) -> float:
        """
        Evaluate the propagation speed of one AP, in m/s

        Parameters
        ----------
        x_AP : NDArray
            x positions of the AP
        t_AP : NDArray
            t positions of the AP

        Returns
        -------
        float
            propagation of the AP, in m/s
        """
        x_start, t_start = self.get_start_AP(x_AP, t_AP)
        # if self.get_AP_upward_len(x_AP,t_AP) > self.get_AP_downward_len(x_AP,t_AP):
        x_stop, t_stop = self.get_xmax_AP(x_AP, t_AP)
        # else:
        #    x_stop,t_stop  = self.get_xmin_AP(x_AP,t_AP)
        speed = np.abs(x_stop - x_start) / (t_stop - t_start)
        return speed * 1e-3

    def remove_raster_artifacts(self, vm_key: str = "V_mem") -> None:  # **kwargs
        """
        Remove artifacts (non-APs) rasterized points.

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"
        """
        x_APs, x_idx, t_APs, t_idx = self.split_APs(vm_key)
        x_clean = []
        x_idx_clean = []
        t_clean = []
        t_idx_clean = []

        for i, _ in enumerate(x_APs):
            x_clean += list(x_APs[i])
            x_idx_clean += list(x_idx[i])
            t_clean += list(t_APs[i])
            t_idx_clean += list(t_idx[i])

        self[vm_key + "_raster_position"] = np.array(x_idx_clean)
        self[vm_key + "_raster_x_position"] = np.array(x_clean)
        self[vm_key + "_raster_time_index"] = np.array(t_idx_clean)
        self[vm_key + "_raster_time"] = np.array(t_clean)

    def rasterize(
        self, vm_key: str = "V_mem", clear_artifacts: bool = True, **kwargs
    ) -> None:
        """
        Rasterize a membrane potential (or filtered or any quantity processed from membrane voltage), with AP detection.
        This function adds 4 items to the class, with the key termination '_raster_position', '_raster_x_position', '_raster_time_index', '_raster_time' concatenated to the original key.
        These keys correspond to:

            - _raster_position    : AP position as the indice of the original key
            - _raster_x_position  : AP position as geometrical position in um
            - _raster_time_index  : AP time as the indice of the original key
            - _raster_time        : AP time as ms

        Parameters
        ----------
        vm_key : str, optional
            Rasterized Vmembrane key , by default "V_mem"

        clear_artifacts : bool
            if True, remove artifacts (non-APs) rasterized points. By default True.

        **kwargs:
            - threshold: (float) threshold for AP detection, in mV. Default value is taken from axon model
            - t_min_AP: (float) minimum AP duration, in ms. Default is 0.1ms
            - t_refractory: (float) inter-AP duration, in ms. Default is 0.5ms
            - t_start: (float) start time for rasterize, in ms. Default is 0ms
            - t_stop: (float) stop time for rasterize, in ms. Default is simulation time.
        """

        #clear_artifacts = False
        if not vm_key + "_raster_position" in self:

            if "threshold" in kwargs:
                threshold = kwargs["threshold"]
            elif "threshold" in self:
                threshold = self["threshold"]
            else:
                threshold = 0

            if "t_min_AP" in kwargs:
                t_min_AP = kwargs["t_min_AP"]
            else:
                t_min_AP = 0.1

            if "t_refractory" in kwargs:
                t_refractory = kwargs["t_refractory"]
            else:
                t_refractory = 0.5

            t_start = None
            if "t_start" in kwargs:
                t_start = kwargs["t_start"]

            t_stop = None
            if "t_stop" in kwargs:
                t_start = kwargs["t_stop"]

            ## selecting the list of position considering what has been recorded
            vm = self[vm_key].copy()
            if self["myelinated"] == True:
                if self["rec"] == "all":
                    x_idx = self["node_index"]
                    x = self["x"][x_idx]
                    vm = vm[x_idx]
                else:
                    x = self["x_rec"]
            else:
                x = self["x_rec"]

            (
                self[vm_key + "_raster_position"],
                self[vm_key + "_raster_x_position"],
                self[vm_key + "_raster_time_index"],
                self[vm_key + "_raster_time"],
            ) = rasterize(
                y_data=vm,
                x_data=x,
                t_data=self.t,
                threshold=threshold,
                t_min_AP=t_min_AP,
                t_refractory=t_refractory,
                t_start=t_start,
                t_stop=t_stop,
            )
            if clear_artifacts:
                self.remove_raster_artifacts(vm_key)

    def speed(self, position_key=None, t_start=0, t_stop=0, x_start=0, x_stop=0):
        """
        Compute the velocity of a spike from rasterized data in a dictionary. The speed can be either positive or negative depending on the propagation direction.

        Parameters
        ----------
        key         : str
            name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
        t_start     : float
            time at which the spikes are processed, in ms. By default 0
        t_stop      : float
            maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
        x_start     : float
            minimum position for spike processing, in um. By default set to 0.
        x_stop      : float
            maximum position for spike processing, in um. If 0 is specified, spikes are processed on the full axon length . By default set to 0.

        Returns
        -------
        speed   : float
            velocity.

        Note
        ----
        the velocity is computed with first and last occurance of a spike, be careful specifying the computation window if multiple spikes. This function will not see velocity variation.
        """

        rise_warning(
            "DeprecationWarning: ",
            "speed is obsolete, please use get_avg_AP_speed() instead",
        )
        # define max timing if not already defined
        if t_stop == 0:
            t_stop = self["t_sim"]
        if t_start == 0:
            if "intra_stim_starts" in self and self["intra_stim_starts"] != []:
                t_start = self["intra_stim_starts"][0]
        if x_start == 0:
            x_stop = self["L"]
        elif x_stop == 0:
            x_start = self["L"]
        # find the best raster plot
        if position_key == None:
            if "V_mem_filtered_raster_position" in self:
                good_key_prefix = "V_mem_filtered_raster"
            elif "V_mem_raster_position" in self:
                good_key_prefix = "V_mem_raster"
            else:
                # there is no rasterized voltage, nothing to find
                return False
        else:
            good_key_prefix = position_key
        # get data only in time windows
        sup_time_indexes = np.where(self[good_key_prefix + "_time"] > t_start)
        inf_time_indexes = np.where(self[good_key_prefix + "_time"] < t_stop)
        good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
        sup_position_indexes = np.where(
            self[good_key_prefix + "_x_position"][good_indexes_time] >= x_start
        )
        inf_position_indexes = np.where(
            self[good_key_prefix + "_x_position"][good_indexes_time] <= x_stop
        )
        good_indexes_position = np.intersect1d(
            sup_position_indexes, inf_position_indexes
        )
        good_indexes = np.intersect1d(good_indexes_position, good_indexes_time)
        good_spike_times = self[good_key_prefix + "_time"][good_indexes]
        good_spike_positions = self[good_key_prefix + "_x_position"][good_indexes]
        max_time = np.argmax(good_spike_times)
        min_time = np.argmin(good_spike_times)
        speed = (
            (good_spike_positions[max_time] - good_spike_positions[min_time])
            * 10**-3
            / (good_spike_times[max_time] - good_spike_times[min_time])
        )
        return speed

    def block(self, position_key=None, t_start=0, t_stop=0):
        """
        check if an axon is blocked or not. The simulation has to include the test spike. This function will look for the test spike initiation and check the propagation

        Parameters
        ----------
        key 		: str
            name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
        t_start		: float
            time at which the test spikes can occur, in ms. By default 0
        t_stop		: float
            maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.

        Returns
        -------
        flag 	: bool or None
            True if the axon is blocked, False if not blocked and None if the test spike does not cross the stimulation near point in the simulation (no possibility to check for the axon state)
        """

        # rise_warning(
        # "DeprecationWarning: ",
        # "block is obsolete, please used isBlocked() instead"
        # )
        if "blocked" in self:
            return self["blocked"]
        position_max = 0
        blocked_spike_positionlist = []
        if t_stop == 0:
            t_stop = self["t_sim"]
        if t_start == 0:
            if "intra_stim_starts" in self and self["intra_stim_starts"] != []:
                t_start = self["intra_stim_starts"][0]

        if position_key == None:
            if "V_mem_filtered_raster_position" in self:
                good_key_prefix = "V_mem_filtered_raster"
            elif "V_mem_raster_position" in self:
                good_key_prefix = "V_mem_raster"
            else:
                # there is no rasterized voltage, nothing to find
                self["blocked"] = False
                return self["blocked"]
        sup_time_indexes = np.where(self[good_key_prefix + "_time"] > t_start)
        inf_time_indexes = np.where(self[good_key_prefix + "_time"] < t_stop)
        good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
        good_spike_times = self[good_key_prefix + "_time"][good_indexes_time]
        blocked_spike_positionlist = self[good_key_prefix + "_x_position"][
            good_indexes_time
        ]
        if len(blocked_spike_positionlist) == 0:
            self["blocked"] = None
            return self["blocked"]
        if "intra_stim_positions" in self:
            if self["intra_stim_positions"] < self["extracellular_electrode_x"]:
                position_max = max_spike_position(
                    blocked_spike_positionlist, position_max, spike_begin="down"
                )

                if blocked_spike_positionlist[position_max] < 9.0 / 10 * self["L"]:
                    self["blocked"] = True
                    return self["blocked"]
                else:
                    for i in range(position_max - 1):
                        if (
                            blocked_spike_positionlist[i + 1]
                            - blocked_spike_positionlist[i]
                            > self["L"] / 5
                        ):
                            self["blocked"] = True
                            return self["blocked"]
                    else:
                        self["blocked"] = False
                        return self["blocked"]
            else:
                position_max = max_spike_position(
                    blocked_spike_positionlist, position_max, spike_begin="up"
                )
                if min(blocked_spike_positionlist) > 1.0 / 10 * self["L"]:
                    self["blocked"] = True
                    return self["blocked"]
                else:
                    for i in range(position_max - 1):
                        if (
                            blocked_spike_positionlist[i]
                            - blocked_spike_positionlist[i + 1]
                            > self["L"] / 5
                        ):
                            self["blocked"] = True
                            return self["blocked"]
                    else:
                        self["blocked"] = False
                        return self["blocked"]
        else:
            rise_error("intra_stim_positions is not in dictionnary")

    def check_test_AP(self):
        """
        Check if a test AP is correctely triggered during an axon simulation and if so return the\
        trigger time

        Parameters
        ----------
        self     : dict
            results of axon.simulate method

        Returns
        -------
        test_AP     : float or None
            time in ms of the first test AP during the simulation. None if no test AP found
        """

        rise_warning("DeprecationWarning: ", "check_test_AP is obsolete")
        if type(self) == str:
            self = self.load_simulation_from_json()
        if "intra_stim_starts" not in self:
            return None
        else:
            mask = False
            test_AP = self["intra_stim_starts"]
            if len(test_AP):
                if np.iterable(test_AP):
                    test_AP = test_AP[0]
                i_first_pos = np.where(self["V_mem_raster_x_position"] == 0)
                for i in i_first_pos[0]:
                    if (
                        self["V_mem_raster_time"][i] >= test_AP - 0.01
                        and self["V_mem_raster_time"][i] <= test_AP + 0.7
                    ):
                        mask = True
                if not mask:
                    test_AP = None
            else:
                test_AP = None
            return test_AP

    def detect_start_extrastim(self, threshold=None):
        """
        Returns the starting time of extracellular stimulation from axon simulation results

        Parameters
        ----------
        self     : dict
            results of axon.simulate method
        threshold       : float
            Current threshold (uA) to consider the stimulation started, if None take the time of second point of the\
            stimulation, by default None

        Returns
        -------
        t_start     : list or float or None
            list of stimulation starting time in ms, float if only one stimulation and None if no stimulation
        """

        rise_warning("DeprecationWarning: ", "detect_start_extrastim is obsolete")
        t_start = []
        if "extracellular_stimuli" in self:
            for i in range(len(self["extracellular_stimuli"])):
                s = self["extracellular_stimuli"][i]
                t = self["extracellular_stimuli_t"][i]
                if threshold is None:
                    t_start += [t[1]]
                else:
                    for i in reversed(range(len(s))):
                        if abs(s[i]) > threshold:
                            t_start += [t[i]]
        if len(t_start) == 0:
            t_start = None
        elif len(t_start) == 1:
            t_start = t_start[0]
        return t_start

    def extra_stim_properties(self)->dict:
        """
        Return elect caracteristics (blocked, Onset response, ...)

        Returns
        -------
        electrode       : dict
            dictonry containing the position (x, y, z), the stimulation start (ms) and maximum value (uA)
            for each electrodes
        """

        electrode = {}

        if "extracellular_electrode_y" in self:
            electrode["x"] = self["extracellular_electrode_x"]
            electrode["y"] = self["extracellular_electrode_y"]
            electrode["z"] = self["extracellular_electrode_z"]
            electrode["start"] = self.detect_start_extrastim(self)
            electrode["max amplitude"] = [max(s) for s in self["extracellular_stimuli"]]

        return electrode

    def block_summary(
        self, AP_start: float, freq: float = None, t_refractory: float = 1
    ) -> dict:
        """
        Return axon block characteristics: blocked, onset response, number of APs)

        Parameters
        ----------
        AP_start : float
            timestamp of the test pulse start, in ms.
        freq : float, optional
            Frequency of the stimulation, for KES block, by default None
        t_refractory : float, optional
            Axon refractory period, by default 1

        Returns
        -------
        axon_state       : dict
            tuple containing the block characteristics: is_blocked (bool), has_onset (bool), n_onset (int)
        """

        if {"is_blocked", "has_onset", "n_onset"} not in self:
            vm_key = "V_mem"
            # if freq is not None: 
            #     vm_key += "_filtered"
            # required_keys = {f"{vm_key}_raster_position",
            #                 f"{vm_key}_raster_x_position",
            #                 f"{vm_key}_raster_time_index",
            #                 f"{vm_key}_raster_time"}
            required_keys = {vm_key}

            if required_keys in self:
                #vm_key = "V_mem"
                #if freq is not None:
                #    self.filter_freq("V_mem", freq, Q=2)
                #    vm_key += "_filtered"

                blocked = self.is_blocked(AP_start, freq, t_refractory)
                n_APs = self.count_APs(vm_key) - 1
                if n_APs < 0:
                    n_APs = 0
                onset = n_APs > 0

                axon_state = {
                    "is_blocked": blocked,
                    "has_onset": onset,
                    "n_onset": n_APs,
                }
                self.update(axon_state)
            else:
                rise_warning(
                    "The following keys are missing.",
                    required_keys - set(self.keys()),
                    " Please check the simulation parameters",
                )
                return None

        else:
            axon_state = {
                "is_blocked": self.is_blocked,
                "has_onset": self.has_onset,
                "n_onset": self.n_onset,
            }
        return axon_state
    
    def find_central_index(self)->int:
        """
        Returns the index of the closer node from the center

        Returns
        -------
        int
            index of `x_rec` of the closer node from the center
        """
        n_center = len(self["x_rec"]) // 2
        return n_center

    ##################################
    ## Impedance properties methods ##
    ##################################

    def compute_f_mem(self):
        """
        compute the cutoff frequency of the axon's membrane and add it to the simulation self dictionnary
        NB: The frequency is computed in [kHz]

        Parameters
        ----------
        self_sim     : dict
            self of axon.simulate method

        Returns
        -------
        f_mem              : np.ndarray
            value of the cutoff frequency of the axon's membrane
        """

        if "g_mem" not in self:
            rise_warning(
                "f_mem cannot be computed computed without membrane conductivity"
            )
            return None
        if "f_mem" not in self:
            ax = self.generate_axon()
            self["c_mem"] = ax.get_membrane_capacitance()
            del ax
        N_x, N_t = np.shape(self["g_mem"])
        f_mem = np.zeros((N_x, N_t))
        for i_t in range(N_t):
            f_mem[:, i_t] = self["g_mem"][:, i_t] / (2 * np.pi * self["c_mem"])

        # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
        # [MHz] to convert to [kHz]
        self["f_mem"] = to_nrv_unit(f_mem, "MHz")
        return self["f_mem"]

    def get_membrane_conductivity(
        self,
        x: float | None = None,
        t: float | None = None,
        i_x: int | np.ndarray | None = None,
        i_t: int | np.ndarray | None = None,
        unit: str = "S/cm**2",
        mem_th: float = 7 * nm,
    ) -> float:
        """
        get the membrane conductivity at a position x and a time t
         
        Parameters
        ----------
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface conductivity in [S]/([m]*[m]): from neuron simulation
            - the conductivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """
        if "g_mem" not in self:
            rise_warning("to get membrane conductivity the ")
            return None

        n_t = len(self["g_mem"][0])
        n_x = len(self["g_mem"][:, 0])
        if t is None:
            if i_t is None:
                i_t = np.arange(n_t)
        else:
            i_t = int(n_t * t / self["t_sim"])
        if x is None:
            if i_x is None:
                i_x = np.arange(n_x)
        else:
            i_x = np.argmin(abs(self["x_rec"] - x))
        
        if np.iterable(i_x) and np.iterable(i_t):
            g = self["g_mem"][np.ix_(i_x, i_t)]
        else:
            g = self["g_mem"][i_x, i_t]
    
        # Surface conductivity in [S]/([m]*[m])
        if "2" in unit:
            return convert(g, "S/cm**2", unit)
        # conductivity in [S]/[m]
        else:
            g *= from_nrv_unit(mem_th, "cm")
            return convert(g, "S/cm", unit)

    def get_membrane_capacitance(
        self, unit: str = "uF/cm**2", mem_th: float = 7 * nm
    ) -> float:
        """
        get the axon membrane capacitance or permitivity

        Parameters
        ----------
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm

        Note
        ----
        depending of the unit parameter this function either return :

            - the surface conductivity in [S]/([m]*[m]): from neuron simulation
            - the conductivity in [S]/[m]:  by multiplying surface conductivity by membrane thickness
        """
        c_mem = membrane_capacitance_from_model(self.model)

        # Surface capacity in [F]/([m]*[m])
        if "2" in unit:
            return convert(c_mem, "S/cm**2", unit)
        # permitivity in [F]/[m]
        else:
            c_mem *= from_nrv_unit(mem_th, "cm")
            return convert(c_mem, "S/cm", unit)

    def get_membrane_complexe_admitance(
        self,
        f: float = 1.0,
        x: float | None = None,
        t: float | None = None,
        i_x: int | np.ndarray | None = None,
        i_t: int | np.ndarray | None = None,
        unit: str = "S/m",
        mem_th: float = 7 * nm,
    ) -> np.array:
        """
        get the membran complexe admitance of each axon at a position x and a time t for a given frequency

        Parameters
        ----------
        f : float or np.array, optional
            effective frequency in kHz, by default 1
        x : float, optional
            x-position in um where to get the conductivity, by default 0
        t : float, optional
            simulation time in ms when to get the conductivity, by default 0
        unit : str, optional
            unit of the returned conductivity see `units`, by default "S/cm**2"
        mem_th : float, optional
            membrane thickness in um, by default 7*nm
        """
        c = self.get_membrane_capacitance(mem_th=mem_th)
        g = self.get_membrane_conductivity(x=x, t=t, i_x=i_x, i_t=i_t, mem_th=mem_th)
        f_mem = g / (2 * np.pi * c)

        # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
        # [MHz] to convert to [kHz]
        f_mem = to_nrv_unit(f_mem, "MHz")

        Y = compute_complex_admitance(f=f, g=g, fc=f_mem)

        if "2" in unit:
            return convert(Y, "S/cm**2", unit)
        # permitivity in [F]/[m]
        else:
            Y *= from_nrv_unit(mem_th, "cm")
            return convert(Y, "S/cm", unit)

    def get_membrane_material(
        self, t: float = 0, unit: str = "S/m", mem_th: float = 7 * nm, **interp_kwargs
    ) -> f_material:

        n_t = len(self["g_mem"][0])
        i_t = int(n_t * t / self["t_sim"])
        x_ = self["x_rec"]
        g_ = deepcopy(self["g_mem"][:, i_t])

        # Surface conductivity in [S]/([m]*[m])
        if "2" in unit:
            g_ = convert(g_, "S/cm**2", unit)
        # conductivity in [S]/[m]
        else:
            g_ *= from_nrv_unit(mem_th, "cm")
            g_ = convert(g_, "S/cm", unit)
        return mat_from_interp(x_, g_, **interp_kwargs)

    #####################
    ## Ploting methods ##
    #####################

    def raster_plot(self, axes: plt.axes, key: str = "V_mem", **kwgs) -> None:
        required_keys = {key, "tstop", "L"}
        if required_keys in self:
            self.rasterize(key)
            axes.scatter(
                self[key + "_raster_time"], self[key + "_raster_x_position"], **kwgs
            )
            axes.set_xlim(0, self["tstop"])
            axes.set_ylim(0, self["L"])
        else:
            rise_warning(
                "The following keys are missing.",
                required_keys - set(self.keys()),
                " Please check the simulation parameters",
            )

    def colormap_plot(self, axes: plt.axes, key: str = "V_mem", switch_axes:bool=False, **kwgs) -> plt.colorbar:
        required_keys = {key, "tstop", "L", "t"}
        if required_keys in self:
            x_ = self["t"]
            y_ = self.get_axon_xrec()
            c_ = self[key]

            if switch_axes:
                x_, y_, c_ = y_, x_,c_.T
                axes.set_ylabel("Time (ms)")
                axes.set_xlabel("Position (µm)")
                axes.set_ylim(0, self["tstop"])
                axes.set_xlim(0, self["L"])
            else:
                axes.set_xlabel("Time (ms)")
                axes.set_ylabel("Position (µm)")
                axes.set_xlim(0, self["tstop"])
                axes.set_ylim(0, self["L"])
            map = axes.pcolormesh(
                x_, y_, c_, shading="auto", **kwgs
            )

            cbar = plt.colorbar(map)
            cbar.set_label(key)
            return cbar
        else:
            rise_warning(
                "The following keys are missing.",
                required_keys - set(self.keys()),
                " Please check the simulation parameters",
            )
        return None

    @abstractmethod
    def plot_x_t(
        self,
        axes: plt.Axes,
        x_index: np.ndarray,
        x_pos: np.ndarray,
        key: str = "V_mem",
        color: str = "k",
        switch_axes=False,
        norm=None,
        **kwgs
    ) -> None:
        """
        Plot `key` in time at various position.

        Warning
        -------
        This method should only internally called by either `myelinated.plot_x_t` or `unmyelinated.plot_x_t`. Therefore, `x_index`, `x_pos` arguments are automatically set by the daughter class methods

        Parameters
        ----------
        axes : plt.Axes
            axes of the figure to display the fascicle
        x_index : np.ndarray
            list of index to plot
        x_pos : np.ndarray
            x position that should be ploted
        key : str, optional
            key of the results' signal that should be ploted, by default "V_mem"
        color : str, optional
            color of the line plot, by default "k"
        switch_axes : bool, optional
            _description_, by default False
        """
        dx = np.abs(x_pos[1] - x_pos[0])
        if norm is not None:
            norm_fac = norm
        else:
            norm_fac = dx / abs(np.max(self[key] - np.min(self[key])) * 1.1)
        offset = np.abs(np.min(self[key][0] * norm_fac))
        for x, x_idx in zip(x_pos, x_index):
            x_ = self["t"]
            y_ = self[key][x_idx] * norm_fac + x + offset
            if switch_axes:
                x_, y_ = y_, x_[::-1]
            axes.plot(
                x_,
                y_,
                color=color, **kwgs
            )



""" nmod Results: standardized result gathering for nmod simulations.

nmod.results defines classes to gather results from nmod object simulations.
Result classes are defined for axons (myelinated and unlyelinated), fascicles
and nerves. All computed values, state variables and electrical quantities are
stored and exportable as dictionaries (json format).

"""

from ._axons_results import axon_results
from ._unmyelinated_results import unmyelinated_results
from ._myelinated_results import myelinated_results
from ._fascicles_results import fascicle_results
from ._nerve_results import nerve_results


submodules = []

classes = [
    "axon_results",
    "unmyelinated_results",
    "myelinated_results",
    "fascicle_results",
    "nerve_results",
]

functions = []

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



""" User Interface - ui: end user pre- or post processing code

ui provides classes and functions for pre- (meshing for instance) or
post-processing computations. Basic or recurent simulation encapuslated
as function are also accessible.

"""


from ._axon_simulations import (
    search_threshold_dispatcher,
    axon_AP_threshold,
    axon_block_threshold,
    firing_threshold_point_source,
    firing_threshold_from_axon,
    blocking_threshold_point_source,
    blocking_threshold_from_axon,
)

from ._axon_postprocessing import (
    remove_key,
    remove_non_NoR_zones,
    generate_axon_from_results,
    filter_freq,
    rasterize,
    AP_detection,
    speed,
    block,
    max_spike_position,
    count_spike,
    check_test_AP,
    detect_start_extrastim,
    extra_stim_properties,
    axon_state,
    get_index_myelinated_sequence,
    find_central_node_index,
    compute_f_mem,
    get_myelin_properties,
    plot_Nav_states,
    default_PP,
    rmv_keys,
    is_recruited,
    is_blocked,
    sample_g_mem,
    vmem_plot,
    raster_plot,
)

from ._fascicle_postprocessing import(
    ls_axons_results,
    ls_csv,
    rm_file,
    rm_sim_dir,
    rm_sim_dir_from_results,
    CAP_time_detection,
    fascicular_state,
    plot_fasc_state,
)

from ._NRV_Msh import (
    mesh_from_electrode,
    mesh_from_extracellular_context,
    mesh_from_fascicle,
    mesh_from_nerve,
)

submodules = []

classes = []

functions = [
    "search_threshold_dispatcher",
    "axon_AP_threshold",
    "axon_block_threshold",
    "firing_threshold_point_source",
    "firing_threshold_from_axon",
    "blocking_threshold_point_source",
    "blocking_threshold_from_axon",
    "remove_key",
    "remove_non_NoR_zones",
    "generate_axon_from_results",
    "filter_freq",
    "rasterize",
    "AP_detection",
    "speed",
    "block",
    "max_spike_position",
    "count_spike",
    "check_test_AP",
    "detect_start_extrastim",
    "extra_stim_properties",
    "axon_state",
    "get_index_myelinated_sequence",
    "find_central_node_index",
    "compute_f_mem",
    "get_myelin_properties",
    "plot_Nav_states",
    "default_PP",
    "rmv_keys",
    "is_recruited",
    "is_blocked",
    "sample_g_mem",
    "vmem_plot",
    "raster_plot",
    "ls_axons_results",
    "ls_csv",
    "rm_file",
    "rm_sim_dir",
    "rm_sim_dir_from_results",
    "CAP_time_detection",
    "fascicular_state",
    "plot_fasc_state",
    "mesh_from_electrode",
    "mesh_from_extracellular_context",
    "mesh_from_fascicle",
    "mesh_from_nerve",
]

__all__ = []

__all__ += submodules
__all__ += classes
__all__ += functions



"""
NRV-:class:`.Optimizer` handling.
"""

from abc import ABCMeta, abstractmethod
import os
from typing import Any
import numpy as np
import matplotlib.pyplot as plt
from time import perf_counter, asctime, localtime
from multiprocessing import cpu_count
import traceback
import json
import sys
import pyswarms as ps
from pyswarms.backend.topology import Star, Ring
import scipy.optimize as scpopt


from ..backend._NRV_Class import NRV_class
from ..backend._parameters import parameters
from ..backend._log_interface import set_log_level, rise_error, rise_warning

from .optim_utils._OptimResults import optim_results

dir_path = parameters.nrv_path + "/_misc"
dir_path + "/log/NRV.log"
# mymodule.py


class Optimizer(NRV_class, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, method=None):
        super().__init__()
        self._method = method
        self.swarm_optimizer = False

    def minimize(self, f: callable, **kwargs) -> optim_results:
        """
        Minimze the function ``f`` using the `optimzer`'s parameters

        Parameters
        ----------
        f : callable
            function to minimize
        kwargs                  : dict
            containing parameters to change in the class (:class:``PSO_optimizer``)

        Returns
        -------
        optim_results
            results of the optimization
        """
        self.set_parameters(**kwargs)
        results = optim_results(self.save(save=False))
        results["date"] = asctime(localtime())
        results["status"] = "Processing"
        return results

    def __call__(self, f: callable, **kwargs: Any) -> optim_results:
        return self.minimize(f, **kwargs)


class scipy_optimizer(Optimizer):
    """


    Parameters
    ----------
    method : str | callable, optional
        Type of solver.  Should be one of
            - 'Nelder-Mead'
            - 'Powell'
            - 'CG'
            - 'BFGS'
            - 'Newton-CG'
            - 'L-BFGS-B'
            - 'TNC'
            - 'COBYLA'
            - 'SLSQP'
            - 'trust-constr'
            - 'dogleg'
            - 'trust-ncg'
            - 'trust-exact'
            - 'trust-krylov'
            - custom - a callable object

    x0 : np.ndarray, optional
        Initial guess. Array of real elements of size (n,), where n is the number of independent variables,. by default None
    args : tuple, optional
        Extra arguments passed to the objective function and its derivatives (fun, jac and hess functions), by default ()
    jac : {callable,  '2-point', '3-point', 'cs', bool}, optional
        Method for computing the gradient vector, optional
        _description_, by default None
    hess : {callable, '2-point', '3-point', 'cs', HessianUpdateStrategy}, optional
        Method for computing the Hessian matrix, by default None
    hessp : callable, optional
        Hessian of objective function times an arbitrary vector p. Only for
        Newton-CG, trust-ncg, trust-krylov, trust-constr, by default None
    bounds : sequence or `Bounds`, optional
        Bounds on variables for Nelder-Mead, L-BFGS-B, TNC, SLSQP, Powell,
        trust-constr, and COBYLA methods., by default None
    constraints : {Constraint, dict} or List of {Constraint, dict}, optional
        Constraints definition. Only for COBYLA, SLSQP and trust-constr, by default ()
    tol : float, optional
        Tolerance for termination, by default None
    callback : callable, optional
        A callable called after each iteration, by default None
    maxiter : int, optional
        _description_, by default None
    options : dict, optional
        A dictionary of solver options, by default None
    dimension : int, optional
        _description_, by default None
    normalize : bool, optional
        _description_, by default False

    Note
    ----
    The following arguments are those used in scipy.optimize.minimize (see scipy documentation [2] for more informations): ``x0``, ``args``, ``jac``, ``hess``, ``hessp``, ``bounds``, ``constraints``, ``tol``, ``callback``, ``options``

    Examples
    --------
    The following lines shows how `scipy_optimizer` can be used to minimze a 2D sphere function.

    >>> import nrv
    >>> my_cost1 = nrv.sphere()
    >>> my_opt = nrv.scipy_optimizer(dimension=2, x0= [100, 10], maxiter=10)
    >>> res = my_opt.minimize(my_cost1)
    >>> print("best position",res.x, "best cost", res.best_cost)

        best position [-1.52666570e-08 -1.32835147e-07] best cost 1.337095622502969e-07


    References
    ----------
    [1] `scipy <https://docs.scipy.org/doc/scipy/>`_

    [2] `scipy.optimize.minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize>`_
    """

    def __init__(
        self,
        method: str = None,
        x0: np.ndarray = None,
        args: tuple = (),
        jac: callable = None,
        hess: callable = None,
        hessp: callable = None,
        bounds: tuple = None,
        constraints: dict = (),
        tol: float = None,
        callback: callable = None,
        maxiter: int = None,
        options: dict = None,
        dimension: int = None,
        normalize: bool = False,
    ):
        if method is None:
            super().__init__("scipy_default")
        else:
            super().__init__("scipy_" + method)
        self.x0 = x0
        self.args = args
        self.scipy_method = method
        self.jac = jac
        self.hess = hess
        self.hessp = hessp
        self.bounds = bounds
        self.constraints = constraints
        self.tol = tol
        self.callback = callback
        self.options = options or {}
        self.maxiter = maxiter
        self.dimensions = dimension
        self.normalize = normalize
        self.scale_translation = None
        self.scale_homothety = None
        self.scaled_bounds = None

    def __update_dimensions(self, results):
        """ """
        if self.bounds is None:
            if self.dimensions is not None:
                self.bounds = [(None, None) for _ in range(self.dimensions)]
        if self.x0 is None:
            if self.bounds is None:
                rise_error(
                    "scipy optimizer: at least x0 or boundaries should be initiate"
                )
            else:
                self.dimensions = len(self.bounds)
                self.x0 = []

                for i in range(self.dimensions):
                    min, max = self.bounds[i]
                    if min is None:
                        min = np.finfo(np.float32).min
                    if max is None:
                        max = np.finfo(np.float32).max
                    self.x0 += [np.random.uniform(min, max)]
                results["x0"] = self.x0
        else:
            self.dimensions = len(self.x0)
            results["dimensions"] = self.dimensions

    def __normalize_bound(self, results):
        self.scale_translation = np.zeros(self.dimensions)
        self.scale_homothety = np.ones(self.dimensions)
        if not self.normalize:
            self.scaled_bounds = self.bounds
        else:
            self.scaled_bounds = []
            for i, bd in enumerate(self.bounds):
                self.scaled_bounds += [(0, 1)]
                self.scale_translation[i] = bd[0]
                self.scale_homothety[i] = bd[1] - bd[0]
        results["scale_translation"] = self.scale_translation
        results["scale_homothety"] = self.scale_homothety
        results["scaled_bounds"] = self.scaled_bounds

    def minimize(self, f: callable, **kwargs) -> optim_results:
        results = super().minimize(f, **kwargs)
        if self.maxiter is not None:
            self.options["maxiter"] = self.maxiter
        self.__update_dimensions(results)
        self.__normalize_bound(results)
        results["cost_history"] = []
        results["position_history"] = []

        def f_history(x):
            x = (results["scale_homothety"] * x) + results["scale_translation"]
            c = f(x)
            results["position_history"].append([x])
            results["cost_history"].append(c)
            return c

        t0 = perf_counter()
        res = scpopt.minimize(
            fun=f_history,
            x0=self.x0,
            args=self.args,
            method=self.scipy_method,
            jac=self.jac,
            hess=self.hess,
            hessp=self.hessp,
            bounds=self.scaled_bounds,
            constraints=self.constraints,
            tol=self.tol,
            callback=self.callback,
            options=self.options,
        )
        res.x = (results["scale_homothety"] * res.x) + results["scale_translation"]

        t_opt = perf_counter() - t0
        results["optimization_time"] = t_opt
        results["best_cost"] = res.fun
        results["best_position"] = res.x

        results.update(res)
        # hess_inv cannot be converted to json
        if "hess_inv" in results:
            del results["hess_inv"]
        results["status"] = "Completed"
        return results


class PSO_optimizer(Optimizer):
    """
    Perform a Particle swarm optimization (PSO) on with a defined cost function using pyswarms
    library[1]

    Parameters
    ----------
    n_particles           	: int
        number of particle of the swarm, by default 5
    dimensions              : int
        number of dimensions of each particle
    options                 : dict
        hyperparameter of the PSO
    maxiter                    : int
        number of iteration of the PSO
    n_processes             : int
        number of process used to parallelize cost calculation, by default None
    bounds                  : tupple
        bounds of the particle, if equal no bounds, by default (0, 0)
    init_pos                : array
        initial position of the particles if None random, by default None
    print_time              : bool
        if True, print the optimisation time, by default True
    opt_type                : str
        Neightboorhood type, by default "global"
        type possibly:

            - "global"                : Global best PSO (star topology)
            - "local"                 : Local best PSO (ring topology)

    static      : bool
        if False and opt_type is local, update the neigthboorhood of each particle every iterations
    bh_strategy             : str
        out of bound position strategy for pyswarms optimizer [2]:

            - "nearest"               : Round the value to the nearest bound (default)
            - "periodic"              : set to the modulus of the value between the two bounds
            - "random"                : set to a random value
            - "shrink"                : reduce the velocity to finish land on the bound
            - "reflective"            : mirror the position form inside to ouside the bounds
            - "intermediate"          : set to intermediate value between previous pos and bound

    oh_strategy             : dict (like {"w":str, "c1":str, "c2":str})
        Dynamic options strategy for pyswarms optimizer [3], if None static options,
        by default None:

            - "exp_decay"             : Decreases the parameter exponentially between limits
            - "lin_variation"         : Decreases/increases the parameter linearly between limits
            - "nonlin_mod"            : Decreases/increases the parameter between limits
                                    according to a nonlinear modulation index
            - "rand"                  : takes a uniform random value between limits

    ftol                    : float
        relative error in objective_func(best_pos) acceptable for convergence, if None -np.inf
        default None
    ftol                    : int
        number of iterations over which the relative error in objective_func(best_pos) is
        acceptable for convergence, by default 1
    save_results            : bool
        save or not the output in a .json file, by default False
    saving_file             : str
        name of the file on wich the output should be saved, by default "pso_results.json"

    Returns
    -------
    results     : optim_results
        contains all the parameters and outputs of the PSO

    References
    ----------
    links to pyswarms doc:
        [1] `Pyswams <https://pyswarms.readthedocs.io/en/latest/index.html>`_

        [2] `Pyswams handlers <https://pyswarms.readthedocs.io/en/latest/api/pyswarms.handlers.html>`_
    """

    def __init__(
        self,
        n_particles=5,
        bounds=(0, 0),
        dimensions=50,
        options=None,
        maxiter=1,
        n_processes=None,
        init_pos=None,
        print_time=False,
        opt_type="global",
        static=False,
        ftol=None,
        ftol_iter=1,
        bh_strategy="nearest",
        oh_strategy=None,
        save_results=False,
        saving_file="pso_results.json",
        comment=None,
    ):
        super().__init__("PSO")
        self.swarm_optimizer = True
        self.n_particles = n_particles
        self.dimensions = dimensions
        self.options = options
        self.maxiter = maxiter
        self.n_processes = n_processes
        self.bounds = bounds
        self.init_pos = init_pos
        self.print_time = print_time
        self.opt_type = opt_type
        self.static = static
        self.ftol = ftol
        self.ftol_iter = ftol_iter
        self.bh_strategy = bh_strategy
        self.oh_strategy = oh_strategy
        self.save_results = save_results
        self.saving_file = saving_file
        self.comment = comment

    def __nrv2pyswarms_bounds(self):
        if np.size(self.bounds) == 2 * self.dimensions:
            if isinstance(self.bounds[0], tuple) and np.size(self.bounds[0]) == 2:
                max_bound = np.zeros(self.dimensions)
                min_bound = np.zeros(self.dimensions)
                for i, bd in enumerate(self.bounds):
                    max_bound[i] = max(bd)
                    min_bound[i] = min(bd)
                return (min_bound, max_bound)
        if np.size(self.bounds) > 2:
            return self.bounds
        elif self.bounds[0] == self.bounds[1]:
            return None
        else:
            max_bound = max(self.bounds) * np.ones(self.dimensions)
            min_bound = min(self.bounds) * np.ones(self.dimensions)
            return (min_bound, max_bound)

    def minimize(self, f_swarm: callable, **kwargs) -> optim_results:
        """
        Perform a Particle swarm optimization

        Parameters
        ----------
        f_swarm     : callable
            function taking in parameter a swarm (dim-dimensions array) and returning the cost for each
            particle (1-dimensionnal array)
        kwargs                  : dict
            containing parameters to change in the class (:class:``PSO_optimizer``)

        Returns
        -------
        optim_results
            results of the optimization
        """
        # self.__mproc_handling()
        results = super().minimize(f_swarm, **kwargs)

        verbose = parameters.get_nrv_verbosity() > 2
        # initial position
        if not np.iterable(self.init_pos):
            self.init_pos = None

        # create pyswarms bounds
        psbounds = self.__nrv2pyswarms_bounds()

        if self.opt_type.lower() == "local":
            if not self.options or len(self.options) < 5:
                self.options = {"c1": 0.5, "c2": 0.5, "w": 0.5, "k": 5, "p": 2}
            topology = Ring(static=self.static)
        # Check number of pcu
        else:
            if not self.options:
                self.options = {"c1": 0.5, "c2": 0.5, "w": 0.5}
            topology = Star()

        # Initialize results directory
        if self.opt_type.lower() != "global":
            results["optimization_parameters"]["static"] = self.static

        if self.comment is not None:
            results["comment"] = self.comment

        if self.save_results:
            with open(self.saving_file, "w") as outfile:
                json.dump(results, outfile)

        if self.ftol is None:
            ## As np.inf cannot be encode by json encoding ftol is used as a local varialble
            ftol = -np.inf
            self.ftol_iter = 1
        else:
            ftol = self.ftol

        try:
            # Call instance of PSO
            optimizer = ps.single.general_optimizer.GeneralOptimizerPSO(
                n_particles=self.n_particles,
                dimensions=self.dimensions,
                options=self.options,
                oh_strategy=self.oh_strategy,
                bounds=psbounds,
                ftol=ftol,
                ftol_iter=self.ftol_iter,
                init_pos=self.init_pos,
                bh_strategy=self.bh_strategy,
                topology=topology,
            )
            optimizer.rep = set_log_level("WARNING", clear_log_file=True)
            # Perform optimization
            t0 = perf_counter()
            cost, pos = optimizer.optimize(
                f_swarm,
                iters=self.maxiter,
                n_processes=self.n_processes,
                verbose=verbose,
            )
            t_opt = perf_counter() - t0

            pos_history = np.array(optimizer.pos_history)
            self.nit = np.shape(pos_history[:, 0, :])[0]

            if self.print_time:
                print(
                    "the cumputing time for ",
                    self.nit,
                    " iterations and ",
                    self.n_particles,
                    " particles is : ",
                    t_opt,
                    "s",
                )

            # Save results
            results["nit"] = self.nit
            results["nfev"] = self.nit * self.n_particles
            results["status"] = "Completed"
            results["best_position"] = pos.tolist()
            results["x"] = results["best_position"]
            results["best_cost"] = cost

            results["optimization_time"] = t_opt
            results["cost_history"] = optimizer.cost_history
            results["neighbor_bestc"] = optimizer.mean_neighbor_history
            results["personal_bestc"] = optimizer.mean_pbest_history

            velocity_history = np.array(optimizer.velocity_history)
            pos_history = np.array(optimizer.pos_history)
            for i in range(self.n_particles):
                results["position" + str(i + 1)] = pos_history[:, i, :].tolist()
                results["velocity" + str(i + 1)] = velocity_history[:, i, :].tolist()
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)
        except KeyboardInterrupt:
            results["status"] = "Interrupted"
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)
            rise_error(KeyboardInterrupt, "Interuption raised during PSO")
        except:
            results["status"] = "Failed"
            results["Error_from_prompt"] = traceback.format_exc()
            print(results["Error_from_prompt"])
            if self.save_results:
                with open(self.saving_file, "w") as outfile:
                    json.dump(results, outfile)

        return results



from ..backend._NRV_Class import NRV_class
from ..backend._log_interface import rise_error


class eit_inverse(NRV_class):
    """
    Class allowing image reconstruction from ``eit_forward`` results. Based on the `pyEIT` algorithms.
    """

    def __init__():
        _protocol = None
        rise_error(NotImplementedError)



"""
Access and modify NRV Parameters.
"""

import configparser
import os
from ._NRV_Singleton import NRV_singleton
from pathlib import Path
from multiprocessing import Process, current_process, parent_process, active_children, Lock

is_master_proc = current_process().name=="MainProcess"

class nrv_parameters(metaclass=NRV_singleton):
    """
    A class for NRV parameters used to gather parameters
    """

    def __init__(self):
        """
        Initialize the class for parameters
        """
        if type(self) in nrv_parameters._instances:
            self = self._instances[type(self)]
        else:
            self.nrv_path = str(Path(os.path.dirname(__file__)).parent.absolute()) 
            self.dir_path = self.nrv_path + "/_misc"
            self.config_fname = self.dir_path + "/NRV.ini"
            self.load()

    def save(self):
        """
        Saving the parameters
        """
        with open(self.config_fname, "w") as configfile:
            self.machine_config.write(configfile)

    def load(self, fname=None):
        """
        Loading the parameters
        """
        if fname is None:
            fname = self.config_fname

        self.machine_config = configparser.ConfigParser()
        self.machine_config.read(self.config_fname)

        # GMSH
        self.GMSH_Ncores = int(self.machine_config.get("GMSH", "GMSH_CPU"))
        self.GMSH_Status = self.machine_config.get("GMSH", "GMSH_STATUS") == "True"
        # nmod
        self.nmod_Ncores = int(self.machine_config.get("NRV", "NMOD_CPU"))
        # optim
        self.optim_Ncores = int(self.machine_config.get("OPTIM", "OPTIM_CPU"))

        # LOG
        self.LOG_Status = self.machine_config.get("LOG", "LOG_STATUS") == "True"
        self.VERBOSITY_LEVEL = int(self.machine_config.get("LOG", "VERBOSITY_LEVEL"))

    def get_nrv_verbosity(self):
        """
        get general verbosity level
        """
        return self.VERBOSITY_LEVEL

    def set_nrv_verbosity(self, i):
        """
        set general verbosity level
        Parameters
        ----------
        O: None
        1: + Error
        2: + Warning
        3: + Info
        4: + Debug
        NB: to add Debug verbosity to the log VERBOSITY_LEVEL has to be set to 4
        """
        self.VERBOSITY_LEVEL = i


    ############################
    # Multi process parameters #
    ############################
    @property
    def is_alone(self):
        return current_process().name=="MainProcess" and len(active_children())==0
    
    @property
    def proc_label(self):
        return current_process().name

    def get_gmsh_ncore(self):
        """
        get gmsh core number
        """
        return self.GMSH_Ncores

    def set_gmsh_ncore(self, n:int):
        """
        set gmsh core number
        """
        self.GMSH_Ncores = n

    def get_nmod_ncore(self):
        """
        get nmod core number
        """
        return self.nmod_Ncores

    def set_nmod_ncore(self, n:int):
        """
        set nmod core number
        """
        self.nmod_Ncores = n

    def get_optim_ncore(self):
        """
        get optim core number
        """
        return self.optim_Ncores

    def set_optim_ncore(self, n:int):
        """
        set optim core number
        """
        self.optim_Ncores = n

    def set_ncores(self, n_nrv:int=None, n_nmod:int=None, n_gmsh:int=None,n_optim:int=None):
        """
        set for all subpakages core number

        Parameters
        ----------
        n_nmod : int, optional
            _description_, by default None
        n_gmsh : int, optional
            _description_, by default None
        n_optim : int, optional
            _description_, by default None
        """
        if n_nrv is not None:
            n_nmod = n_nrv
            n_gmsh = n_nrv
            n_optim = n_nrv
        if n_nmod is not None:
            self.set_nmod_ncore(n_nmod)
        if n_gmsh is not None:
            self.set_gmsh_ncore(n_gmsh)
        if n_optim is not None:
            self.set_optim_ncore(n_optim)


###########################
#   Parameter singleton   #
###########################

parameters = nrv_parameters()
